#  Copyright (C) 2012 by Carnegie Mellon University.
#  
#  @OPENSOURCE_HEADER_START@
#  Use of the Rayon and related source code is subject to the terms of
#  the following licenses:
#  
#  GNU Public License (GPL) Rights pursuant to Version 2, June 1991
#  Government Purpose License Rights (GPLR) pursuant to DFARS 252.227.7013
#  
#  NO WARRANTY
#  
#  ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER 
#  PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY 
#  PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN 
#  "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY 
#  KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT 
#  LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE, 
#  MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE 
#  OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT, 
#  SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY 
#  TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF 
#  WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES. 
#  LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF 
#  CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON 
#  CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE 
#  DELIVERABLES UNDER THIS LICENSE.
#  
#  Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie 
#  Mellon University, its trustees, officers, employees, and agents from 
#  all claims or demands made against them (and any related losses, 
#  expenses, or attorney's fees) arising out of, or relating to Licensee's 
#  and/or its sub licensees' negligent use or willful misuse of or 
#  negligent conduct or willful misconduct regarding the Software, 
#  facilities, or other rights or assistance granted by Carnegie Mellon 
#  University under this License, including, but not limited to, any 
#  claims of product liability, personal injury, death, damage to 
#  property, or violation of any laws or regulations.
#  
#  Carnegie Mellon University Software Engineering Institute authored 
#  documents are sponsored by the U.S. Department of Defense under 
#  Contract FA8721-05-C-0003. Carnegie Mellon University retains 
#  copyrights in all material produced under this contract. The U.S. 
#  Government retains a non-exclusive, royalty-free license to publish or 
#  reproduce these documents, or allow others to do so, for U.S. 
#  Government purposes only pursuant to the copyright license under the 
#  contract clause at 252.227.7013.
#  @OPENSOURCE_HEADER_END@
from __future__ import division
import wx

import isilk_rwuniq
import isilk_rwcount

from rayon import data, datatypes, toolbox, RayonException
from netsa import json, times
from netsa_silk import IPAddr

class RayonIsilkException(RayonException):
    pass


class Info(object):
    def __init__(self, data, metadata):
        self.data = data
        self.metadata = metadata

class RayonIsilkPanel(wx.Panel):
    """Base class for Rayon visualization panels used in iSiLK."""
    # Subclasses should supply:
    #
    #   chartmod: the module containing get_chart
    #
    #   metadata_defaults: self-explanatory
    #
    # Subclasses should implement:
    #
    #   munge_data: to turn the iSiLK data into the data
    #   chartmod.get_chart expects
    chartmod = None
    @classmethod
    def munge_metadata(cls, input_metadata):
        # Make default metadata dictionary, then update it with new data
        metadata = dict(cls.metadata_defaults)
        metadata.update(input_metadata)
        return metadata
        
    def __init__(self, data, metadata=None, viewpage=None, **wx_panelargs):
        wx.Panel.__init__(self, **wx_panelargs)

        if metadata is None:
            metadata = {}

        # View page is the thing we call methods on when we're clicked
        # on and whatnot
        self.viewpage = viewpage

        # This is my half-baked attempt to changing data in the object
        # atomic. To really work, it probably needs to hold the chart,
        # too, but that's even sillier.
        self.info = Info(self.munge_data(data), self.munge_metadata(metadata))

        self.chart = None

        self.tools = toolbox.Toolbox.for_backend("wx")
        chart = self.chartmod.get_chart(self.tools,
                                        self.info.data,
                                        self.info.metadata)
        self.set_chart(chart)
        self.chartmod.set_events(chart, self)
        
        self.dirty = True
        self.buffer = None
        self.Bind(wx.EVT_PAINT, self._evt_paint)
        self.Bind(wx.EVT_SIZE, self._evt_size)
        self.Bind(wx.EVT_IDLE, self._evt_idle)

    def get_metadata(self):
        return self.metadata

    def set_dirty(self, dirty):
        self.dirty = dirty


    def set_chart(self, chart):
        if self.chart is not None:
            old_event_adapter = self.chart.get_event_source(self)
            old_event_adapter.deactivate()
        self.chart = chart
        new_event_adapter = self.chart.get_event_source(self)
        new_event_adapter.activate()
        self.dirty = True
        self.redraw()
        self.Refresh()
        self.Update()

    def reset_chart(self, new_data, new_metadata=None):
        if new_metadata is None:
            new_metadata = {}
        new_info = Info(self.munge_data(new_data),
                        self.munge_metadata(new_metadata))
        chart = self.chartmod.get_chart(self.tools,
                                        new_info.data,
                                        new_info.metadata)
        self.info = new_info
        self.set_chart(chart)
        

    def redraw(self):
        if self.chart is None:
            self.buffer = wx.EmptyBitmap(1, 1)
        else:
            size = self.GetClientSize()
            self.buffer = wx.EmptyBitmap(max(1, size.width),
                                         max(1, size.height))
            dc = wx.BufferedDC(None, self.buffer)
            ctx = self.tools.get_backend().Context(dc)
            self.chart(ctx)
            self.dirty = False



    def _evt_size(self, evt):
        # Set the dirty bit. When we stop, _evt_idle will repaint.
        self.dirty = True
        evt.Skip()


    def _evt_idle(self, evt):
        if self.dirty:
            self.redraw()
            # Kick off an EVT_PAINT
            self.Refresh(False)
        # Let others handle evt_idle if they choose
        evt.Skip()

    def _evt_paint(self, evt):
        # Create a buffered paint DC.  It will create the real
        # wx.PaintDC and then blit the bitmap to it when dc is
        # deleted.  Since we don't need to draw anything else
        # here that's all there is to it.
        if self.buffer is not None:
            dc = wx.BufferedPaintDC(self, self.buffer)
            del dc
        evt.Skip()


    # Methods iSiLK expects of me
    def RefreshContents(self):
        pass



class RayonRwuniqPanel(RayonIsilkPanel):
    chartmod = isilk_rwuniq
    metadata_defaults = {
        'show': ['bytes'],
    }
    # Possible output columns in rwuniq, ordered by "importance"
    # (display priority)
    ordered_columns = [('sip'      , IPAddr),
                       ('sport'    , int),
                       ('stype'    , int),
                       ('scc'      , str),
                       ('sval'     , str),
                       ('dip'      , IPAddr),
                       ('dport'    , int),
                       ('dtype'    , int),
                       ('dcc'      , str),
                       ('dval'     , str),
                       ('nhip'     , int),
                       ('protocol' , int),
                       ('itype'    , int),
                       ('icode'    , int),
                       ('sensor'   , str),
                       ('type'     , str),
                       ('class'    , str)]
    # Unordered output columns in rwuniq which have non-string types
    unordered_columns = {
        # TODO: put fields in here that shouldn't just be strings
        'bytes'   : int,
        'packets' : int,
        'records' : int,
       }
    column_order = zip(*ordered_columns)[0]
    column_types = dict(ordered_columns)
    column_types.update(unordered_columns)


    @classmethod
    def dataset_from_datacache(cls, datacache):
        # datacache is a SimpleDataCache object object, as defined in
        # isilk/datacache.py
        keys = set(datacache.GetFieldNames())
        colnames = []
        # Do the columns whose order we care about first...
        for col, coltype in cls.ordered_columns:
            if col in keys:
                colnames.append(col)
                keys.remove(col)
        # ...followed by the unordered columns
        for col, coltype in cls.unordered_columns.items():
            if col in keys:
                colnames.append(col)
                keys.remove(col)
        # Transform them into a list of data in column-major order,
        # and away we go
        dset = [[cls.column_types.get(str(cname), str)(r.getfield(cname))
                 for r in datacache]
                for cname in colnames]
        return data.Dataset.from_columns(dset, colnames=colnames)
    
    def munge_data(self, datacache):
        # Get the raw dataset from the datacache. This dataset
        # contains three "volume columns": 'bytes', 'packets',
        # 'records'. The rest of the columns form the key by which the
        # data was aggregated.
        dset = self.dataset_from_datacache(datacache)

        # The columns in the dataset containing quantitative
        # information, not just categories. (TODO: Right now, iSiLK
        # vis only supports a single value column being displayed at a
        # time.)
        try:
            val_column_idx = [idx for idx, c in enumerate(dset.column_names())
                              if c in ('bytes', 'packets', 'records')][0]
        except IndexError:
            raise RayonIsilkException("Invalid rwuniq output: "
                                      "no bytes, packets or records columns")

        # iSiLK currently takes only the top 10 values from this set,
        # sorted by the value column. This will change to be user
        # settable at some point.
        dset.sort(lambda r: r[val_column_idx])
        dset = dset.filter(lambda r: True, limit=10)
        
        

        # Now chop it up into two smaller datasets.
        #
        # table_data contains the first five columns in the key (These
        # will be displayed in the table to the left of the vis)
        #
        # bar_data contains 4 columns:
        #   "key": a multicolumn containing 1all the columns in the key
        #   "bytes": byte volume
        #   "packets": packet volume
        #   "records": record volume
        val_cols = []
        val_colnames = []
        key_cols = []
        key_colnames = []
        for cname in dset.column_names(): 
            if cname in ('bytes', 'packets', 'records'):
                val_cols.append(dset.column(cname))
                val_colnames.append(cname)
            else:
                key_cols.append(dset.column(cname))
                key_colnames.append(cname)
        key_data = data.Dataset.from_columns(key_cols, colnames=key_colnames)
        val_data = data.Dataset.from_columns(val_cols, colnames=val_colnames)
        table_data = data.Dataset.from_columns(key_cols[:5],
                                               colnames=key_colnames[:5])
        return key_data.meld(), val_data, table_data
        
            

class RayonRwcountPanel(RayonIsilkPanel):
   chartmod = isilk_rwcount
   metadata_defaults = {
       'show': ['bytes'],
   }

   column_types = {
       'date'    : times.make_datetime,
       'bytes'   : float,
       'packets' : float,
       'records' : float,
   }
   
   def munge_data(self, datacache):
       # The datacache contains time series data. The time column is
       # labeled 'date'; the remainder of the columns -- 'bytes',
       # 'packets' and 'records' -- are volume columns.
       keys = set(datacache.GetFieldNames())
       colnames = [col for col in self.column_types.keys()
                   if col in keys]
       
       #colnames = [name for name in datacache.GetFieldNames()]
       dset = [[self.column_types.get(str(cname), str)(r.getfield(cname))
                for r in datacache]
               for cname in colnames]
       return data.Dataset.from_columns(dset, colnames=colnames)
        
__all__ = ["RayonRwuniqPanel", "RayonRwcountPanel"]
