#  Copyright (C) 2012 by Carnegie Mellon University.
#  
#  @OPENSOURCE_HEADER_START@
#  Use of the Rayon and related source code is subject to the terms of
#  the following licenses:
#  
#  GNU Public License (GPL) Rights pursuant to Version 2, June 1991
#  Government Purpose License Rights (GPLR) pursuant to DFARS 252.227.7013
#  
#  NO WARRANTY
#  
#  ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER 
#  PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY 
#  PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN 
#  "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY 
#  KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT 
#  LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE, 
#  MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE 
#  OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT, 
#  SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY 
#  TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF 
#  WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES. 
#  LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF 
#  CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON 
#  CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE 
#  DELIVERABLES UNDER THIS LICENSE.
#  
#  Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie 
#  Mellon University, its trustees, officers, employees, and agents from 
#  all claims or demands made against them (and any related losses, 
#  expenses, or attorney's fees) arising out of, or relating to Licensee's 
#  and/or its sub licensees' negligent use or willful misuse of or 
#  negligent conduct or willful misconduct regarding the Software, 
#  facilities, or other rights or assistance granted by Carnegie Mellon 
#  University under this License, including, but not limited to, any 
#  claims of product liability, personal injury, death, damage to 
#  property, or violation of any laws or regulations.
#  
#  Carnegie Mellon University Software Engineering Institute authored 
#  documents are sponsored by the U.S. Department of Defense under 
#  Contract FA8721-05-C-0003. Carnegie Mellon University retains 
#  copyrights in all material produced under this contract. The U.S. 
#  Government retains a non-exclusive, royalty-free license to publish or 
#  reproduce these documents, or allow others to do so, for U.S. 
#  Government purposes only pursuant to the copyright license under the 
#  contract clause at 252.227.7013.
#  @OPENSOURCE_HEADER_END@
from __future__ import division
from rayon import RayonException
import math
import itertools

from core import *

from rayon.geometry import Rectangle, Spacing, ScreenPoint
from rayon.subdivide import subdivide_htiles, subdivide_vtiles

class SimpleLayout(Layout):
    """
    A layout with a single component (not a class:`Layout` object,
    but a component with a ``draw`` method.), rendered in the center.
    """
    @classmethod
    def wrap(cls, layout_or_panel):
        # Utility method. If layout_or_panel can be laid out, returns
        # it; otherwise, returns a SimpleLayout wrapping
        # layout_or_panel.
        try:
            if layout_or_panel.is_layoutable():
                return layout_or_panel
            else:
                raise AttributeError()
        except AttributeError:
            return cls(layout_or_panel, BoxSpacing(0, 0, 0, 0))
        
    def __init__(self, component, spacing):
        Layout.__init__(self, spacing)
        if component is not None and not hasattr(component, 'draw'):
            raise RayonException("Child component of SimpleLayout ('%s') "
                                 "does not support draw() method" % 
                                 component.__class__.__name__)
        self.child = component
    def get_component(self): return self.child
    def layout_(self, ctx, topleft, width=None, height=None, call_depth=0):
        spc = self.spacing
        width, height = ctx.normalize_dimensions(width, height)

        padded_topleft = spc.padded_topleft(ctx, topleft, width, height)
        padded_bottomright = spc.padded_bottomright(
            ctx, point_add(topleft, (width, height)), width, height)

        ctx.push_state()
        try:
            ctx.translate(spc.left_padding(ctx, width),
                          spc.top_padding(ctx, height))
            # None is allowed as a child. We simply do nothing with it.
            if self.child is not None:
                self.child.draw(ctx,
                                spc.inner_width(ctx, width),
                                spc.inner_height(ctx, height))
            # No children, so...
            return [], (padded_topleft, padded_bottomright), True
        finally:
            ctx.pop_state()
            


class FloatLayout(Layout):
    """A layout containing a single L{Layout} child, floating
    in the C{FloatLayout} box at a specified C{x, y} coordinate. """
    def __init__(self, layout_or_panel, x, y, width, height):
        """Constructor. Width and height of the float box are
        determined by the spacing of the child layout.

        The C{FloatLayout} does not take a spacing argument. The
        position, width and height of the float are all independent of
        padding.

        @type  layout_or_panel: L{Layout} or L{rayon.base.Panel}
        @param layout_or_panel: A child layout or panel that will be
            rendered as a float box.

        @type  x: length specification
        @param x: position of left edge of float box on X axis, as a
            proportion of the total width

        @type  y: length specification
        @param y: position of top edge of float box on Y axis, as a
            proportion of the total height.

        @type  width: length specification
        @param width: width of float box
        
        @type  height: length specification
        @param height: height of float box
        """
        Layout.__init__(self, BoxSpacing(0, 0, 0, 0))
        self.child = SimpleLayout.wrap(layout_or_panel)
        self.x = units_from_spec(x)
        self.y = units_from_spec(y)
        self.box_width = units_from_spec(width)
        self.box_height = units_from_spec(height)
        # TODO: make sure the float doesn't overflow the box and/or
        # padding...maybe
    def layout_(self, ctx, topleft, width=None, height=None, call_depth=0):
        width, height = ctx.normalize_dimensions(width, height)
        x = self.x.to_pixels(ctx, width)
        y = self.y.to_pixels(ctx, height)
        box_width = self.box_width.to_pixels(ctx, width)
        box_height = self.box_height.to_pixels(ctx, height)

        ctx.push_state()
        try:
            ctx.translate(*(x, y))
            grandkids, kid_bbox, is_leaf = self.child.layout(
                ctx, point_add(topleft, (x, y)), box_width, box_height,
                call_depth=call_depth+1)
            childnode = LayoutNode(grandkids, kid_bbox[0], kid_bbox[1], 0,
                                   self.child, is_leaf)
            # We only draw on the part of the canvas that our child
            # drew on, so...
            return [childnode], kid_bbox, False
        finally:
            ctx.pop_state()


class LayeredLayout(Layout):
    """A layout containing multiple L{Layout}s, which it draws
    one on top of the other."""
    def __init__(self, spacing):
        Layout.__init__(self, spacing)
        self.layouts = []
        self.layouts_by_name = {}
        self.layouts_rlookup = {}
    def add_layout(self, layout_or_panel, name=None):
        layout = SimpleLayout.wrap(layout_or_panel)
        self.layouts.append(layout)
        if name is not None:
            if name in self.layouts_by_name:
                raise RayonException("Cannot add two children to the same "
                                   "component with the same name: "
                                   "'%s'" % name)
            self.layouts_by_name[name] = layout
            self.layouts_rlookup[layout] = name
            layout.name = name
        else:
            layout.name = None
        return layout
    def get_layout(self, name):
        return self.layouts_by_name.get(name, None)
    def get_layout_name(self, layout):
        return self.layouts_rlookup[layout]
    def layouts_iter(self):
        return (l for l in self.layouts)
    def layout_(self, ctx, topleft, width=None, height=None, call_depth=0):
        width, height = ctx.normalize_dimensions(width, height)
        spc = self.spacing
        offset = spc.left_padding(ctx, width), spc.top_padding(ctx, height)
        ctx.push_state()
        try:
            ctx.translate(*offset)
            childnodes = []
            for idx, lay in enumerate(self.layouts):
                try:
                    grandkids, kid_bbox, is_leaf = lay.layout(
                        ctx, 
                        point_add(topleft, offset),
                        spc.inner_width(ctx, width), 
                        spc.inner_height(ctx, height),
                        call_depth=call_depth+1)
                except RayonRenderException, e:
                    try:
                        lay_name = self.layouts_rlookup[lay]
                    except KeyError:
                        raise e
                    else:
                       raise e.add_context((lay.__class__.__name__, lay_name))
                childnodes.append(LayoutNode(
                    grandkids, kid_bbox[0], kid_bbox[1], idx, lay, is_leaf))
            real_topleft = spc.padded_topleft(ctx, topleft, width, height)
            real_bottomright = spc.padded_bottomright(
                ctx, point_add(topleft, (width, height)), width, height)
            return childnodes, (real_topleft, real_bottomright), False
        finally:
            ctx.pop_state()


class GridLayout(Layout):
    def __init__(self, spacing, col_weights, row_weights, fill):
        # All arguments are required.
        
        # *spacing* is a :class:`BoxSpacing` object.

        # *col_weights* is an iterable of integer column weights.

        # *row_weights* it an iterable of integer row weights.

        # *fill* is the string ``horizontal`` or ``vertical``. A
        # horizontal 3x3 grid will be laid out in this order:
        
        #    0|1|2
        #    3|4|5
        #    6|7|8

        # A vertical grid will be laid out in this order:

        #    0|3|6
        #    1|4|7
        #    2|5|8
        self.col_weights = list(col_weights)
        self.row_weights = list(row_weights)
        if len(self.row_weights) == 0 and len(self.col_weights) == 0:
            raise RayonException("Must specify at least one of "
                               "row or column weights")
        self.spacing = spacing
        self.layouts = []
        self.layouts_by_name = {}
        self.layouts_rlookup = {}
        if fill == 'horizontal':
            self.subdivide_fn = subdivide_htiles
        elif fill == 'vertical':
            self.subdivide_fn = subdivide_vtiles
        else:
            raise RayonException("Fill value must be 'horizontal' "
                               "or 'vertical'")
    def has_fixed_layout(self):
        "Return True if the layout has a fixed # of both rows and columns."
        return not (self.col_weights == [] or self.row_weights == [])
    def num_columns(self):
        if len(self.col_weights) == 0:
            return int(math.ceil(len(self.layouts) / len(self.row_weights)))
        else:
            return len(self.col_weights)
    def num_rows(self):
        if len(self.row_weights) == 0:
            return int(math.ceil(len(self.layouts) / len(self.col_weights)))
        else:
            return len(self.row_weights)
    def add_layout(self, layout_or_panel, name=None):
        if name is not None:
            if name in self.layouts_by_name:
                raise RayonException("Multiple named components "
                                     "with same name '%s'" % name)
        layout = SimpleLayout.wrap(layout_or_panel)
        if self.has_fixed_layout():
            if len(self.layouts) + 1 > (self.num_columns() * self.num_rows()):
                raise RayonException("Too many children for GridLayout")
        self.layouts.append(layout)
        if name is not None:
            self.layouts_by_name[name] = layout
            self.layouts_rlookup[layout] = name
    def get_layout(self, name):
        return self.layouts_by_name.get(name, None)
    def layouts_iter(self):
        return (l for l in self.layouts)

    def layout_(self, ctx, topleft, width=None, height=None, call_depth=0):
        width, height = ctx.normalize_dimensions(width, height)

        abs_tl = ScreenPoint(*topleft)

        spc = Spacing(*self.spacing.get_all_padding(ctx, width, height))
        rect, rel_tl = Rectangle(width, height).pad(spc)
        rel_br = rect.get_corners()[3]

        subrects_iter = self.subdivide_fn(
                rect, self.col_weights, self.row_weights)

        def layout_child(lay, subrect, l_tl):
            w, h = subrect.get_width(), subrect.get_height()
            x, y = (rel_tl + l_tl)
            abs_x, abs_y = (abs_tl + rel_tl + l_tl)
            ctx.push_state()
            try:
                ctx.translate(x, y)
                try:
                    grandkids, kid_bbox, is_leaf = lay.layout(
                        ctx, (abs_x, abs_x), w, h, call_depth=call_depth+1)
                except RayonRenderException, e: 
                    try:
                        lay_name = self.layouts_rlookup[lay]
                    except KeyError:
                        raise e
                    else:
                        raise e.add_context((lay.__class__.__name__, lay_name))
                return LayoutNode(
                    grandkids, kid_bbox[0], kid_bbox[1], 0, lay, is_leaf)
            finally:
                ctx.pop_state()
            
        
        childnodes = [layout_child(l, l_rect, l_topleft)    
                      for l, (l_rect, l_topleft)
                      in itertools.izip(self.layouts, subrects_iter)]

        return childnodes, ((abs_tl + rel_tl), (abs_tl + rel_br)), False
                
                

                    

class SquareLayout(Layout):
    """A layout with a main area in the center; borders on the left,
    right, top and bottom; and corners in the northwest, northeast,
    southeast and southwest of a square surface. Instances of
    L{Layout} can be put in any of these positions, providing a
    great deal of flexibility. Additional C{Layout}s may be
    used to lay out the chart background, which is rendered before
    (and therefore \"below\") anything else.

    To those familiar with Java, this class is somewhere between a
    C{BorderLayout} and a C{GridBagLayout}. Except for the background,
    everything added to this layout will be laid out with their
    preferred dimensions except the center plot and the corners, which
    will be laid out with whatever is left.
    """
    def __init__(self, spacing):
        Layout.__init__(self, spacing)
        self.center  = None
        self.borders = dict(left=[], right=[], top=[], bottom=[])
        self.corners = dict(nw=None, ne=None, se=None, sw=None)

    # Borders
    # TODO: What if the margin exceeds the available space?
    def add_left_border(self, layout, width):
        """ Add a L{Layout} along the left edge of the chart."""
        self.borders['left'].append((layout, width))
    def add_right_border(self, layout, width):
        """ Add a L{Layout} along the right edge of the chart."""
        self.borders['right'].append((layout, width))
    def add_top_border(self, layout, height):
        """ Add a L{Layout} along the top edge of the chart."""
        self.borders['top'].append((layout, height))
    def add_bottom_border(self, layout, height):
        """ Add a L{Layout} along the bottom edge of the chart."""
        self.borders['bottom'].append((layout, height))

    # Corners
    def set_corner(self, corner, layout_or_panel):
        self.corners[corner] = SimpleLayout.wrap(layout_or_panel)
    def set_nw_corner(self, layout_or_panel):
        self.set_corner('nw', layout_or_panel)
    def set_ne_corner(self, layout_or_panel):
        self.set_corner('ne', layout_or_panel)
    def set_se_corner(self, layout_or_panel):
        self.set_corner('se', layout_or_panel)
    def set_sw_corner(self, layout_or_panel):
        self.set_corner('sw', layout_or_panel)

    def get_nw_corner(self): return self.corners['nw']
    def get_ne_corner(self): return self.corners['ne']
    def get_se_corner(self): return self.corners['se']
    def get_sw_corner(self): return self.corners['sw']

    # Center
    def set_center(self, layout_or_panel):
        """ Set a layout in the center of the C{SquareLayout}. """
        self.center = SimpleLayout.wrap(layout_or_panel)
        return layout_or_panel

    def get_center(self): return self.center



    # Some commonly-used calculations
    def margin_left(self, ctx, width):
        return sum(w.to_pixels(ctx, width) for b, w in self.borders['left'])
    def margin_right(self, ctx, width):
        return sum(w.to_pixels(ctx, width) for b, w in self.borders['right'])
    def margin_top(self, ctx, height):
        return sum(h.to_pixels(ctx, height) for b, h in self.borders['top'])
    def margin_bottom(self, ctx, height):
        return sum(h.to_pixels(ctx, height) for b, h in self.borders['bottom'])


    def layout_(self, ctx, topleft, width=None, height=None, call_depth=0):

        # xXx debug
        #def p(msg):
        #    print "%s<< %s >>" % (" " * (call_depth + 1), msg)
        def p(msg): pass

        width, height = ctx.normalize_dimensions(width, height)
        my_spc = self.spacing

        # width of drawable space (minus padding)
        inner_width = my_spc.inner_width(ctx, width)
        # height of drawable space (minus padding)
        inner_height = my_spc.inner_height(ctx, height)

        # top padding
        tpadding = my_spc.top_padding(ctx, height)
        # bottom padding
        bpadding = my_spc.bottom_padding(ctx, height)
        # left padding
        lpadding = my_spc.left_padding(ctx, width)
        # right padding
        rpadding = my_spc.right_padding(ctx, width)

        # left and right padding combined
        xpadding = lpadding + rpadding
        # top and bottom padding combined
        ypadding = tpadding + bpadding


        # width of left margin
        lmargin = self.margin_left(ctx, width)
        # width of right margin
        rmargin = self.margin_right(ctx, width)
        # height of top margin
        tmargin = self.margin_top(ctx, height)
        # height of bottom margin
        bmargin = self.margin_bottom(ctx, height)

        # left and right margin combined
        xmargin = lmargin + rmargin
        # top and bottom margin combined
        ymargin = tmargin + bmargin

        # What we'll return at the end of this mess
        childnodes = []
            

        # The corners
        def layout_section(l, tl_x, tl_y, childwidth, childheight):
            # l: Layout object to layout
            # tl_x: x coord of top left corner
            # tl_y: y coord of top left corner
            # childwidth: width of the corner
            # childheight: height of the corner
            ctx.push_state()
            try:
                ctx.translate(tl_x, tl_y, fixup=True)
                grandkids, kid_bbox, is_leaf = l.layout(
                    ctx, point_add(topleft, (tl_x, tl_y)),
                    childwidth, childheight,
                    call_depth=call_depth+1)
                return LayoutNode(grandkids, kid_bbox[0], kid_bbox[1], 0, l,
                                  is_leaf)
            finally:
                ctx.pop_state()    

        # --Northwest
        if self.corners['nw'] is not None:
            corner = self.corners['nw']
            corner_spc = corner.get_spacing()
            corner_width = lmargin
            corner_height = tmargin
            corner_lpad = corner_spc.left_padding(ctx, corner_width)
            corner_tpad = corner_spc.top_padding(ctx, corner_height)
            tl_x = lpadding + corner_lpad
            tl_y = tpadding + corner_tpad
            
            childnodes.append(layout_section(
                    corner, tl_x, tl_y, corner_width, corner_height))

            
        # --Northeast
        if self.corners['ne'] is not None:
            corner = self.corners['ne']
            corner_spc = corner.get_spacing()
            corner_width = rmargin
            corner_height = tmargin
            corner_lpad = corner_spc.left_padding(ctx, corner_width)
            corner_tpad = corner_spc.top_padding(ctx, corner_height)
            tl_x = (width - rpadding - corner_width + corner_lpad)
            tl_y = tpadding + corner_tpad
            
            childnodes.append(layout_section(
                    corner, tl_x, tl_y, corner_width, corner_height))

            
        # --Southeast
        if self.corners['se'] is not None:
            corner = self.corners['se']
            corner_spc = corner.get_spacing()
            corner_width = rmargin
            corner_height = bmargin
            corner_lpad = corner_spc.left_padding(ctx, corner_width)
            corner_tpad = corner_spc.top_padding(ctx, corner_height)
            tl_x = (width - rpadding - corner_width + corner_lpad)
            tl_y = (height - bpadding - corner_height + corner_tpad)

            childnodes.append(layout_section(
                    corner, tl_x, tl_y, corner_width, corner_height))

            
            
        # --Southwest
        if self.corners['sw'] is not None:
            corner = self.corners['sw']
            corner_spc = corner.get_spacing()
            corner_width = lmargin
            corner_height = bmargin
            corner_lpad = corner_spc.left_padding(ctx, corner_width)
            corner_tpad = corner_spc.top_padding(ctx, corner_height)
            tl_x = lpadding + corner_lpad
            tl_y = (height - bpadding - corner_height + corner_tpad)

            childnodes.append(layout_section(
                    corner, tl_x, tl_y, corner_width, corner_height))
            
        # The center
        if self.center is not None:
            center = self.center
            center_spc = center.get_spacing()
            center_width = width - xpadding - xmargin
            center_height = height - ypadding - ymargin
            center_lpad = center_spc.left_padding(ctx, center_width)
            center_tpad = center_spc.top_padding(ctx, center_height)
            tl_x = lpadding + lmargin + center_lpad
            tl_y = tpadding + tmargin + center_tpad

            childnodes.append(layout_section(
                    center, tl_x, tl_y, center_width, center_height))


        # The borders. Borders are drawn from the inside out.

        # Left and right borders
        bheight = height - ypadding - ymargin
        base_tl_y = tpadding + tmargin

        # --Left border(s), right to left
        dx = lpadding + lmargin
        for border, bwidth in self.borders['left']:
            bwidth = bwidth.to_pixels(ctx, width)
            border_spc = border.get_spacing()
            border_tpad = border_spc.top_padding(ctx, bheight)
            border_lpad = border_spc.left_padding(ctx, bwidth)
            tl_x = dx - bwidth + border_lpad
            tl_y = base_tl_y + border_tpad
            childnodes.append(layout_section(
                    border, tl_x, tl_y, bwidth, bheight))
            dx -= bwidth

        # --Right border(s), left to right
        dx = width - rpadding - rmargin
        for border, bwidth in self.borders['right']:
            bwidth = bwidth.to_pixels(ctx, width)
            border_spc = border.get_spacing()
            border_tpad = border_spc.top_padding(ctx, bheight)
            border_lpad = border_spc.left_padding(ctx, bwidth)
            tl_x = dx + border_lpad
            tl_y = base_tl_y + border_tpad
            childnodes.append(layout_section(
                    border, tl_x, tl_y, bwidth, bheight))
            dx += bwidth


        # Top and bottom borders
        bwidth = width - xpadding - xmargin
        base_tl_x = lpadding + lmargin

        # --Top border(s), bottom up
        dy = tpadding + tmargin
        for border, bheight in self.borders['top']:
            bheight = bheight.to_pixels(ctx, height)
            border_spc = border.get_spacing()
            border_tpad = border_spc.top_padding(ctx, bheight)
            border_lpad = border_spc.left_padding(ctx, bwidth)
            tl_y = dy - bheight + border_tpad
            tl_x = base_tl_x + border_lpad
            childnodes.append(layout_section(
                    border, tl_x, tl_y, bwidth, bheight))
            dy -= bheight

            
        # --Bottom border(s), top down
        dy = height - bpadding - bmargin
        for border, bheight in self.borders['bottom']:
            bheight = bheight.to_pixels(ctx, height)
            border_spc = border.get_spacing()
            border_tpad = border_spc.top_padding(ctx, bheight)
            border_lpad = border_spc.left_padding(ctx, bwidth)
            tl_y = dy + border_tpad
            tl_x = base_tl_x + border_lpad
            childnodes.append(layout_section(
                    border, tl_x, tl_y, bwidth, bheight))
            dy += bheight

        # Whew!
        padded_topleft = my_spc.padded_topleft(ctx, topleft, width, height)
        padded_bottomright = my_spc.padded_bottomright(
            ctx, point_add(topleft, (width, height)), width, height)
        p("%s: returning childnodes, (%s, %s), False" % (
                type(self).__name__, padded_topleft, padded_bottomright))
        return childnodes, (padded_topleft, padded_bottomright), False
