#  Copyright (C) 2012 by Carnegie Mellon University.
#  
#  @OPENSOURCE_HEADER_START@
#  Use of the Rayon and related source code is subject to the terms of
#  the following licenses:
#  
#  GNU Public License (GPL) Rights pursuant to Version 2, June 1991
#  Government Purpose License Rights (GPLR) pursuant to DFARS 252.227.7013
#  
#  NO WARRANTY
#  
#  ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER 
#  PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY 
#  PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN 
#  "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY 
#  KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT 
#  LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE, 
#  MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE 
#  OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT, 
#  SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY 
#  TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF 
#  WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES. 
#  LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF 
#  CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON 
#  CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE 
#  DELIVERABLES UNDER THIS LICENSE.
#  
#  Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie 
#  Mellon University, its trustees, officers, employees, and agents from 
#  all claims or demands made against them (and any related losses, 
#  expenses, or attorney's fees) arising out of, or relating to Licensee's 
#  and/or its sub licensees' negligent use or willful misuse of or 
#  negligent conduct or willful misconduct regarding the Software, 
#  facilities, or other rights or assistance granted by Carnegie Mellon 
#  University under this License, including, but not limited to, any 
#  claims of product liability, personal injury, death, damage to 
#  property, or violation of any laws or regulations.
#  
#  Carnegie Mellon University Software Engineering Institute authored 
#  documents are sponsored by the U.S. Department of Defense under 
#  Contract FA8721-05-C-0003. Carnegie Mellon University retains 
#  copyrights in all material produced under this contract. The U.S. 
#  Government retains a non-exclusive, royalty-free license to publish or 
#  reproduce these documents, or allow others to do so, for U.S. 
#  Government purposes only pursuant to the copyright license under the 
#  contract clause at 252.227.7013.
#  @OPENSOURCE_HEADER_END@
import miscutils

# Parsing line specifications


def parse_linespec_string(spec):
    spec_tuple = spec.split(" ")
    if len(spec_tuple) < 3:
        raise RuntimeError("Too few linespec values: %s" % spec)
    if len(spec_tuple) > 3:
        raise RuntimeError("Too many linespec values: %s" % spec)
    return (int(spec_tuple[0]),
            spec_tuple[1],
            miscutils.parse_color(spec_tuple[2]))

def parse_linespec(spec):
    # Take tuple or css-ish "1 solid #fefefe" and turn it into a
    # linespec tuple
    try:
        len(spec)
    except TypeError:
        try:
            wtf = str(spec)
        except:
            wtf = "[Can't represent as string]"
            raise RuntimeError("Unparseable line specification: %s" % wtf)
    if type(spec) == type(""):
        return parse_linespec_string(spec)
    else:
        # Assume it's a sequence in the right order
        return (spec[0], spec[1], miscutils.parse_color(spec[2]))
            
            

# Line Classes

class Line(object):
    @classmethod
    def generic(cls, color, width):
        return cls(color, width)
    
    def __init__(self, color, width, fixup, dashes, offset):
        self.color = miscutils.parse_color(color)
        self.width = width
        self.fixup = fixup
        self.dashes = dashes
        self.offset = offset
    def draw(self, ctx, from_x, from_y, to_x, to_y):
        ctx.draw_line(from_x, from_y, to_x, to_y,
                      self.color, self.width, self.fixup,
                      self.dashes, self.offset)
    def __str__(self):
        return "<lines.Line color: %s; width=%s>" % (self.color, self.width)

class Solid(Line):
    """
    Renders a solid line.

    *color* is the color of the line.

    *width* is the width of the line in pixels/points.

    If *fixup* is ``True``, Rayon will attempt to "fix up" perfectly
    horizontal and vertical lines to avoid anti-aliasing. This
    behavior is disabled if *fixup* is ``False``.
    """
    def __init__(self, color=(0,0,0), width=1, fixup=False):
        Line.__init__(self, color, width, fixup, None, None)
    def __str__(self):
        return "<lines.Solid color: %s; width=%s>" % (self.color, self.width)

class Dashed(Line):
    # Documented in classref-lines.rst
    def __init__(self, color=(0,0,0), width=1, fixup=False):
        Line.__init__(self, color, width, fixup, [5.0, 5.0], 0)
    def __str__(self):
        return "<lines.Dashed color: %s; width=%s>" % (self.color, self.width)

class Dotted(Line):
    # Documented in classref-lines.rst
    def __init__(self, color=(0,0,0), width=1, fixup=False):
        Line.__init__(self, color, width, fixup, [2.0, 5.0], 0)
    def __str__(self):
        return "<lines.Dashed color: %s; width=%s>" % (self.color, self.width)

class DotDash(Line):
    """
    """
    def __init__(self, color=(0,0,0), width=1, fixup=False):
        Line.__init__(self, color, width, fixup, [2, 5, 7, 5], 0)
    def __str__(self):
        return "<lines.Dashed color: %s; width=%s>" % (self.color, self.width)


class Null(Line):
    """
    Can be used like a line, but doesn't actually draw a
    line. *color*, *width* and *fixup* may be passed when this line is
    created, but they are ignored.
    """
    def __init__(self, color=(0,0,0), width=1, fixup=False):
        pass
    def draw(self, ctx, from_x, from_y, to_x, to_y):
        pass
    def __str__(self): return "<lines.Null>"


_line_classes = {
    'solid'   : Solid,
    'dashed'  : Dashed,
    'dotted'  : Dotted,
    'dotdash' : DotDash,
    'none'    : Null, 
}



# Factories

def create(line_style, **line_params):
    try:
        return _line_classes[line_style](**line_params)
    except KeyError, e:
        raise SyntaxError("No such line style %s" % str(e))

def create_generic(line_style, color, width):
    try:
        return _line_classes[line_style].generic(color, width)
    except KeyError, e:
        raise SyntaxError("No such line style %s" % str(e))
    

def create_from_spec(spec, fixup=False):
    width, style, color = parse_linespec(spec)
    return create(style, color, width, fixup)

def list_line_types():
    return tuple((k, v) for k, v in _line_classes.items())

