#  Copyright (C) 2012 by Carnegie Mellon University.
#  
#  @OPENSOURCE_HEADER_START@
#  Use of the Rayon and related source code is subject to the terms of
#  the following licenses:
#  
#  GNU Public License (GPL) Rights pursuant to Version 2, June 1991
#  Government Purpose License Rights (GPLR) pursuant to DFARS 252.227.7013
#  
#  NO WARRANTY
#  
#  ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER 
#  PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY 
#  PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN 
#  "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY 
#  KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT 
#  LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE, 
#  MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE 
#  OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT, 
#  SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY 
#  TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF 
#  WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES. 
#  LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF 
#  CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON 
#  CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE 
#  DELIVERABLES UNDER THIS LICENSE.
#  
#  Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie 
#  Mellon University, its trustees, officers, employees, and agents from 
#  all claims or demands made against them (and any related losses, 
#  expenses, or attorney's fees) arising out of, or relating to Licensee's 
#  and/or its sub licensees' negligent use or willful misuse of or 
#  negligent conduct or willful misconduct regarding the Software, 
#  facilities, or other rights or assistance granted by Carnegie Mellon 
#  University under this License, including, but not limited to, any 
#  claims of product liability, personal injury, death, damage to 
#  property, or violation of any laws or regulations.
#  
#  Carnegie Mellon University Software Engineering Institute authored 
#  documents are sponsored by the U.S. Department of Defense under 
#  Contract FA8721-05-C-0003. Carnegie Mellon University retains 
#  copyrights in all material produced under this contract. The U.S. 
#  Government retains a non-exclusive, royalty-free license to publish or 
#  reproduce these documents, or allow others to do so, for U.S. 
#  Government purposes only pursuant to the copyright license under the 
#  contract clause at 252.227.7013.
#  @OPENSOURCE_HEADER_END@

"""
A set of internally-used utility functions for doing various network
operations. (E.g., IP address conversion.)
"""

from __future__ import division
import struct
import socket


def inet_atoi(addr):
    """
    Returns an integer representation of the given dotted-quad string
    IPv4 address.
    """
    return int(struct.unpack('!L', socket.inet_aton(addr))[0])


def inet_itoa(num):
    """
    Returns a dotted-quad string representation of the given integer
    IPv4 address
    """
    return socket.inet_ntoa(struct.pack('!L', num))



def range_from_network(network, cidr):
    """
    Takes a network address (as an integer from 0-(2^32 - 1)) and a
    CIDR netmask (as an integer from 0-32) and converts them to an IP
    range, returning a tuple of integers representing the minimum and
    maximum IPv4 addresses in the range.

    This function will only output valid CIDR ranges. If the network
    address is not the first address in the CIDR range, the range of
    the specified size that contains the network address will be
    returned. (E.g., passing 0xC0A8010D (192.168.1.13) and a CIDR
    netmask of 24 will yield the range ``(0xC0A80100, 0xC0A801FF)``
    (192.168.1.0-192.168.1.255).
    """
    if not (0 <= cidr <= 32):
        raise ValueError("%d is not a valid CIDR mask")

    inv_prefix = ((2**32 - 1) >> cidr) & 0xFFFFFFFF
    prefix = (~inv_prefix & 0xFFFFFFFF)
    min_ip = network & prefix
    max_ip = min_ip | inv_prefix

    return min_ip, max_ip


def range_from_cidr(ipv4_cidr):
    """
    Converts string representation of an IPv4 CIDR network address to
    an integer and CIDR netmask, then calls :func:`range_from_network`
    to convert to a range.
    """
    network_str, cidr_str = ipv4_cidr.split("/")
    if "." in network_str:
        network  = inet_atoi(network_str)
    else:
        network = int(network_str)
    cidr = int(cidr_str)

    return range_from_network(network, cidr)
