#  Copyright (C) 2012 by Carnegie Mellon University.
#  
#  @OPENSOURCE_HEADER_START@
#  Use of the Rayon and related source code is subject to the terms of
#  the following licenses:
#  
#  GNU Public License (GPL) Rights pursuant to Version 2, June 1991
#  Government Purpose License Rights (GPLR) pursuant to DFARS 252.227.7013
#  
#  NO WARRANTY
#  
#  ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER 
#  PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY 
#  PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN 
#  "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY 
#  KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT 
#  LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE, 
#  MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE 
#  OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT, 
#  SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY 
#  TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF 
#  WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES. 
#  LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF 
#  CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON 
#  CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE 
#  DELIVERABLES UNDER THIS LICENSE.
#  
#  Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie 
#  Mellon University, its trustees, officers, employees, and agents from 
#  all claims or demands made against them (and any related losses, 
#  expenses, or attorney's fees) arising out of, or relating to Licensee's 
#  and/or its sub licensees' negligent use or willful misuse of or 
#  negligent conduct or willful misconduct regarding the Software, 
#  facilities, or other rights or assistance granted by Carnegie Mellon 
#  University under this License, including, but not limited to, any 
#  claims of product liability, personal injury, death, damage to 
#  property, or violation of any laws or regulations.
#  
#  Carnegie Mellon University Software Engineering Institute authored 
#  documents are sponsored by the U.S. Department of Defense under 
#  Contract FA8721-05-C-0003. Carnegie Mellon University retains 
#  copyrights in all material produced under this contract. The U.S. 
#  Government retains a non-exclusive, royalty-free license to publish or 
#  reproduce these documents, or allow others to do so, for U.S. 
#  Government purposes only pursuant to the copyright license under the 
#  contract clause at 252.227.7013.
#  @OPENSOURCE_HEADER_END@

from __future__ import division

from rayon.geometry import ScreenPoint, Rectangle

"""
The :class:`Subdivision` object, and functions for subdividing the
space in :class:`~rayon.geometry.Rectangle` objects.

These functions take a single :class:`~rayon.geometry.Rectangle`
object and return an iterator over a stream of :class:`Subdivision`
objects representing an area of space within the larger
rectangle. Depending on the nature of the subdivision, these regions
may overlap partially or entirely. The iterator may also be an
infinite stream of subdivisions.
"""


class Subdivision(object):
    """
    A subsection of a larger rectangular space, measured relative to
    that larger space. A :class:`Subdivision` object contains a
    :class:`~rayon.geometry.Rectangle` object (call it ``rect``),
    representing the shape of the subdivision and a
    :class:`~rayon.geometry.ScreenPoint` object (call it ``pt``)
    representing the location of the top-left corner of ``rect``,
    relative to the enclosing rectangle.

    The constituent parts of the subdivision can be accessed as in a
    sequence:

       # s is a Subdivision object
       rect = s[0]
       pt = s[1]
       
       # or

       rect, pt = s
    """
    def __init__(self, rect, pt):
        if not isinstance(rect, Rectangle):
            raise TypeError(
                "First argument to Subdivision() must be a Rectangle")
        if not isinstance(pt, ScreenPoint):
            raise TypeError(
                "Second argument to Subdivision() must be a ScreenPoint") 
        self._rect = rect
        self._pt = pt
    def __getitem__(self, key):
        if   key == 0: return self._rect
        elif key == 1: return self._pt
        else: raise IndexError(key)
    def __eq__(self, other):
        return (isinstance(other, Subdivision) and
                self._rect == other._rect and
                self._pt == other._pt)
    def __repr__(self):
        return "Subdivision(%s, %s)" % (self._rect, self._pt)


def _normalize_weights(weights, size):
    if len(weights) == 0:
        return (size,)
    denom = sum(weights)
    return tuple((num / denom) * size for num in weights)
    
    
def subdivide_htiles(inrect, col_weights, row_weights):
    """
    Subdivides *inrect* into a grid, returning a stream that flows
    horizontally through each subdivision.

    *col_weights* and *row_weights* are sequences of integers
    indicating how many rows/columns are in the grid, and what
    proportion of the total space each should occupy. A *col_weight*
    of ``[1,1,1]`` specifies three columns, with each taking up equal
    room. (``[2,2,2]``, ``[3,3,3]``, etc. would be equivalent.) A
    *col_weight* of ``[1,2,3]`` also specifies three columns, but the
    second should take up twice as much space as the first, and the
    third should take up three times as much space as the first.  An
    empty list of weights is equivalent to ``[1]``.

    A weight of 0 for a column or row will cause this function to emit
    a subdivision for cells in that column/row with zero weight --
    that is, with zero width (for columns) or height (for rows). It is
    up to the calling code to decide how to deal with this.
    
    Subdivisions will be returned by iterating "horizontally" -- first
    through rows, then through columns. For example, a grid with
    *col_weights* of ``[1,1,1]`` and *row_weights* of ``[1,1,1]``
    would return cells of a 3x3 grid in the following order::
    
       |0|1|2|
       |3|4|5|
       |6|7|8|

    A weight of 0 will cause this function to emit a subdivision for
    cells in that column/row with zero weight -- that is, with zero
    width (for columns) or height (for rows). It is up to the calling
    code to decide how to deal with this.

    An empty list of weights is equivalent to ``[1]``.
    """
    col_widths = _normalize_weights(tuple(col_weights), inrect.get_width())
    row_heights = _normalize_weights(tuple(row_weights), inrect.get_height())

    x, y = 0, 0
    for h in row_heights:
        for w in col_widths:
            yield Subdivision(Rectangle(w, h), ScreenPoint(x, y))
            x += w
        x = 0
        y += h

def subdivide_vtiles(inrect, col_weights, row_weights):
    """
    Subdivides *inrect* into a grid, returning a stream that flows
    vertically through each subdivision.

    *col_weights* and *row_weights* are sequences of integers
    indicating how many rows/columns are in the grid, and what
    proportion of the total space each should occupy. A *col_weight*
    of ``[1,1,1]`` specifies three columns, with each taking up equal
    room. (``[2,2,2]``, ``[3,3,3]``, etc. would be equivalent.) A
    *col_weight* of ``[1,2,3]`` also specifies three columns, but the
    second should take up twice as much space as the first, and the
    third should take up three times as much space as the first.  An
    empty list of weights is equivalent to ``[1]``.

    A weight of 0 for a column or row will cause this function to emit
    a subdivision for cells in that column/row with zero weight --
    that is, with zero width (for columns) or height (for rows). It is
    up to the calling code to decide how to deal with this.
    
    Subdivisions will be returned by iterating "vertically" -- first
    through columns, then through rows. For example, a grid with
    *col_weights* of ``[1,1,1]`` and *row_weights* of ``[1,1,1]``
    would return cells of a 3x3 grid in the following order::
    
       |0|3|6|
       |1|4|7|
       |2|5|8|
       
    """
    col_widths = _normalize_weights(tuple(col_weights), inrect.get_width())
    row_heights = _normalize_weights(tuple(row_weights), inrect.get_height())

    x, y = 0, 0
    for w in col_widths:
        for h in row_heights:
            yield Subdivision(Rectangle(w, h), ScreenPoint(x, y))
            y += h
        y = 0
        x += w

        

__all__ = """
Subdivision
subdivide_htiles
subdivide_vtiles
""".split()
