import unittest
from StringIO import StringIO


from rayon.dataparse import *
# NOTE: These are not publicly exported
from rayon.dataparse import (
    PushbackReader,
    TypeConversionException,
    _decipher_column_names,
    _decipher_typemap,
    _decipher_delimiter,
    _read_comments,
    _read_header,
)

# NOTE: rayon.dataparse is an internal-only API used by rayon.data to
# marshall and unmarshall serialized datasets.

def dedent(bigstring):
    sio = StringIO(bigstring)
    return "".join(line.lstrip()[1:] for line in sio)


class t_PushbackReader(unittest.TestCase):
    in_str = """|# a comment
                |a line
                |# another comment
                |another line"""
    def test_pushback_1(self):
        pbr = PushbackReader(StringIO(dedent(self.in_str)))
        self.assertEqual("# a comment\n", pbr.readline())
        pbr.pushback()
        self.assertEqual("# a comment\n", pbr.readline())

    def test_peek_1(self):
        pbr = PushbackReader(StringIO(dedent(self.in_str)))
        self.assertEqual("# a comment\n", pbr.peek())
        self.assertEqual("# a comment\n", pbr.peek())

    def test_readline_skipping_1(self):
        pbr = PushbackReader(StringIO(dedent(self.in_str)))
        self.assertEqual(
            "a line\n",
            pbr.readline(skipping=(lambda l: l.startswith('#'),)))

    def test_line_iter_1(self):
        pbr = PushbackReader(StringIO(dedent(self.in_str)))
        self.assertEqual(
            ("a line\n", "another line"),
            tuple(l for l
                  in pbr.line_iter(skipping=(lambda l: l.startswith('#'),))))
        
            
class t_decipher(unittest.TestCase):
    class _(object):
        @staticmethod
        def passes(fn, output):
            def test(self):
                self.assertEquals(output, fn())
            return test
        @staticmethod
        def fails(fn, exc):
            def test(self):
                self.assertRaises(exc, fn)
            return test

    ##################################################
    # _decipher_column_names
    ##################################################


    # user_colnames (from the command-line) is preferred
    test_decipher_column_names_1_1 = _.passes(
        lambda: _decipher_column_names(('a', 'b', 'c'), None, None, "|"),
        ('a', 'b', 'c'))

    test_decipher_column_names_1_2 = _.passes(
        lambda: _decipher_column_names(('a', 'b', 'c'), "a|b|d", "a|b|e", "|"),
        ('a', 'b', 'c'))

    # header colnames is preferred after user_colnames
    test_decipher_column_names_2_1 = _.passes(
        lambda: _decipher_column_names(None, "a|b|d", "a|b|e", "|"),
        ('a', 'b', 'd'))

    test_decipher_column_names_2_2 = _.passes(
        lambda: _decipher_column_names(None, "a,b,d", None, ","),
        ('a', 'b', 'd'))

    test_decipher_column_names_2_3 = _.passes(
        lambda: _decipher_column_names(None, "a,b,d", "a,b,e", ","),
        ('a', 'b', 'd'))

    # Yes, this passes---if there is no delimiter in the row, it's
    # considered one big row. Note, however, the potential for error
    # here. It might be nice to warn about possible delimiter
    # mismatches in the future.
    test_decipher_column_names_2_4 = _.passes(
        lambda: _decipher_column_names(None, "a,b,c", "a,b,e", "|"),
        ("a,b,c",))

    test_decipher_column_names_2_5 = _.passes(
        lambda: _decipher_column_names(None, "a,b,d", "a,b,e", None),
        ("a,b,d",))

    # This works too. It might not even be wrong---the "old-style"
    # column names spec might have just been a normal comment line.
    test_decipher_column_names_2_6 = _.passes(
        lambda: _decipher_column_names(None, "a,b,d", "a|b|e", ","),
        ('a', 'b', 'd'))

    # The old style of specifying column names in a comment right
    # before the dataset is still supported, as a last resort.
    test_decipher_column_names_3_1 = _.passes(
        lambda: _decipher_column_names(None, None, "a|b|e", "|"),
        ('a', 'b', 'e'))

    test_decipher_column_names_3_2 = _.passes(
        lambda: _decipher_column_names(None, None, "a,b,e", ","),
        ('a', 'b', 'e'))

    # Again, the putative old-style column spec might be a plain ol'
    # comment.
    test_decipher_column_names_3_3_1 = _.passes(
        lambda: _decipher_column_names(None, None, "a,b,e", "|"),
        None)

    # Same, even though it could be interpreted as a single column
    # name.
    test_decipher_column_names_3_3_2 = _.passes(
        lambda: _decipher_column_names(None, None, "a,b,e", None),
        None)

    # Whitespace surrounding column names is removed
    test_decipher_column_names_4_1 = _.passes(
        lambda: _decipher_column_names((' a', 'b ', ' c '), None, None, "|"),
        ('a', 'b', 'c'))
        
    test_decipher_column_names_4_2 = _.passes(
        lambda: _decipher_column_names(None, " a|b | e ", None, "|"),
        ('a', 'b', 'e'))
        
    test_decipher_column_names_4_3 = _.passes(
        lambda: _decipher_column_names(None, None, " a|b | e ", "|"),
        ('a', 'b', 'e'))

    # If no names are supplied, return None
    test_decipher_column_names_5_1 = _.passes(
        lambda: _decipher_column_names(None, None, None, None),
        None)

    # (even if we supply a delimiter)
    test_decipher_column_names_5_2 = _.passes(
        lambda: _decipher_column_names(None, None, None, "|"),
        None)
        


    
    ##################################################
    # _decipher_typemap
    ##################################################

    # user_typemap is preferred
    test_decipher_typemap_1_1 = _.passes(
        lambda: _decipher_typemap((int, str), None, None, None),
        (int, str))

    test_decipher_typemap_1_2 = _.passes(
        lambda: _decipher_typemap((int, str), "str|int", None, None),
        (int, str))

    test_decipher_typemap_1_3 = _.passes(
        lambda: _decipher_typemap((int, str), "str|int", "a|b|c", "|"),
        (int, str))

    # hdr_typemap is preferred after user_typemap
    test_decipher_typemap_2_1_1 = _.passes(
        lambda: _decipher_typemap(None, "str|int", None, "|"),
        (str, int))

    test_decipher_typemap_2_1_2 = _.passes(
        lambda: _decipher_typemap(None, "str|int", "a|b|c", "|"),
        (str, int))

    # If delimiter is absent or doesn't occur in the string, the
    # hdr_typemap will be interpreted as a single column
    test_decipher_typemap_2_3_1 = _.passes(
        lambda: _decipher_typemap(None, "int", "a|b|c", None),
        (int,))

    test_decipher_typemap_2_3_2 = _.passes(
        lambda: _decipher_typemap(None, "int", "a|b|c", "|"),
        (int,))


    # As a last resort, types are inferred from the first row
    test_decipher_typemap_2_4_1 = _.passes(
        lambda: _decipher_typemap(None, None, "a|2|c", "|"),
        (str,int,str))

    # If there is no or a bad delimiter, the row is interpreted as a
    # single column
    test_decipher_typemap_2_4_1 = _.passes(
        lambda: _decipher_typemap(None, None, "a|2|c", None),
        (str,))

    test_decipher_typemap_2_4_2 = _.passes(
        lambda: _decipher_typemap(None, None, "a|2|c", ","),
        (str,))

    # If there is no row (and no other way to infer types), an empty
    # tuple is returned. (Which will work fine on an empty dataset,
    # after all.)
    test_decipher_typemap_2_5_1 = _.passes(
        lambda: _decipher_typemap(None, None, None, ","),
        tuple())

    test_decipher_typemap_2_5_2 = _.passes(
        lambda: _decipher_typemap(None, None, None, None),
        tuple())

    ##################################################
    # _decipher_delimiter
    ##################################################

    # user delim is preferred
    test_decipher_delimiter_1_1 = _.passes(
        lambda: _decipher_delimiter(",", None),
        ",")
    
    test_decipher_delimiter_1_2 = _.passes(
        lambda: _decipher_delimiter(",", ";"),
        ",")

    # header delimiter is used if user delim is not supplied
    test_decipher_delimiter_2 = _.passes(
        lambda: _decipher_delimiter(None, ";"),
        ";")

    # default is "|"
    test_decipher_delimiter_3 = _.passes(
        lambda: _decipher_delimiter(None, None),
        "|")


class t_read(unittest.TestCase):
    class _(object):
        @staticmethod
        def passes(fn, str_in, expected):
            def test(self):
                s = PushbackReader(StringIO(dedent(str_in)))
                self.assertEqual(expected, fn(s))
            return test
        @staticmethod
        def next_line_is(fn, str_in, expected_next):
            def test(self):
                s = PushbackReader(StringIO(dedent(str_in)))
                fn(s)
                self.assertEqual(expected_next, s.readline())
            return test
        
    
    ##################################################
    # _read_header
    ##################################################

    # empty values
    test_read_header_0_1 = _.passes(
        _read_header, "", {})

        
    test_read_header_1 = _.passes(
        _read_header,
        """|## foo: bar
           |## baz: quux""",
        {'foo': 'bar', 'baz': 'quux'})

    # keys are downcased. case in values is preserved
    test_read_header_2 = _.passes(
        _read_header,
        """|## Foo: bar
           |## baz: Quux""",
        {'foo': 'bar', 'baz': 'Quux'})

    # No type conversions, implicit or otherwise, are performed. Keys
    # and values are always strings.
    test_read_header_3 = _.passes(
        _read_header,
        """|## foo: 12
           |## baz: quux""",
        {'foo': '12', 'baz': 'quux'})


    # Header parsing ends when the last header comment line is
    # found. No more data is consumed.
    test_read_header_4_0 = _.next_line_is( # ...including empty headers
        _read_header,
        "|# no header here",
        "# no header here")

    test_read_header_4_1 = _.next_line_is(
        _read_header,
        """|## Foo: bar
           |## Baz: quux
           |# this is a random comment""", # note no newline at EOF
        "# this is a random comment")
        
    test_read_header_4_2 = _.next_line_is(
        _read_header,
        """|## Foo: bar
           |## Baz: quux
           |alpha|beta    |gamma
           |delta|episilon|stuff""",
        "alpha|beta    |gamma\n")

        
        
    ##################################################
    # _read_comments
    ##################################################

    test_read_comments_0 = _.passes(
        _read_comments, "", tuple())

    test_read_comments_1 = _.passes(
        _read_comments,
        """|#This is my comment.
           |#There are many like it, but this one is mine.""",
        ("This is my comment.",
         "There are many like it, but this one is mine."))

    # Whitespace between the comment character and the line is ignored
    test_read_comments_2 = _.passes(
        _read_comments,
        """|# This is my comment.
           |# There are many like it, but this one is mine.""",
        ("This is my comment.",
         "There are many like it, but this one is mine."))

    # Additional comment characters are ignored. (So this function
    # will gobble up header lines if you let it. That's desirable most
    # of the time, since a "header line" outside the header is just a
    # comment.)
    test_read_comments_3 = _.passes(
        _read_comments,
        """|## This is my comment.
           |### There are many like it, but this one is mine.""",
        ("This is my comment.",
         "There are many like it, but this one is mine."))

    
    # Parsing ends when a non-comment line is encountered
    test_read_comments_4_0 = _.next_line_is(
        _read_comments,
        "|no|comments",
        "no|comments")

    test_read_comments_4_1 = _.next_line_is(
        _read_comments,
        """|# this is a comment
           |# and so is this
           |this|isn't""",
        "this|isn't")

    # comment lines with leading space ARE NOT comment
    # lines. read_comments will stop when it reaches them, and they
    # will be interpreted as data.
    #
    # This is because they might BE data. For instance:
    #
    #  # symbol | name
    #    #      | octothorpe
    #    ?      | question mark
    test_read_comments_5 = _.passes(
        _read_comments,
        """|# This is a comment
           |  #data that starts with # | stuff """,
        ("This is a comment",))

class t_from_stream(unittest.TestCase):
    class _(object):
        @staticmethod
        def passes(str_in, header, colnames, numcols, rows,
                   orig_colnames=None,
                   orig_typemap=None,
                   orig_delimiter=None,
                   first_line_is_colnames=False,
                   debug=False):
            def test_(self):
                if debug:
                    import pdb; pdb.set_trace()
                p = from_stream(StringIO(dedent(str_in)),
                                orig_colnames,
                                orig_typemap,
                                orig_delimiter,
                                first_line_is_colnames)
                self.assertEqual(header, p.header())
                self.assertEqual(colnames, p.colnames())
                self.assertEqual(numcols, p.numcols())
                self.assertEqual(rows, tuple(tuple(r) for r in p.row_iter()))
            return test_
        @staticmethod
        def fails(str_in, exc,
                  orig_colnames=None,
                  orig_typemap=None,
                  orig_delimiter=None,
                  first_line_is_colnames=False,
                  iterate_rows=True):
            def test_(self):
                def t():
                    p =  from_stream(StringIO(dedent(str_in)),
                                     orig_colnames,
                                     orig_typemap,
                                     orig_delimiter,
                                     first_line_is_colnames)
                    if iterate_rows:
                        tuple(tuple(r) for r in p.row_iter())
                self.assertRaises(exc, t)
            return test_

        # Used in test_missing_4
        @staticmethod
        def missing_typemap(x):
            if x == '': return 'NaN'
            else      : return int(x)


    ## Basic assembly

    test_1 =  _.passes(
        """|# foo | bar | baz | quux
           |    1 | a   | z   | alpha
           |    2 | b   | y   | beta
           |    3 | c   | x   | gamma""",
        {},
        ("foo", "bar", "baz", "quux"),
        4,
        ((1, 'a', 'z', 'alpha'),
         (2, 'b', 'y', 'beta'),
         (3, 'c', 'x', 'gamma')))



    ## Ignoring Comments

    test_comments_1 = _.passes(
        """|# foo | bar | baz | quux
           |    1 | a   | z   | alpha
           |# this comment will be ignored
           |    2 | b   | y   | beta
           |    3 | c   | x   | gamma""",
        {},
        ("foo", "bar", "baz", "quux"),
        4,
        ((1, 'a', 'z', 'alpha'),
         (2, 'b', 'y', 'beta'),
         (3, 'c', 'x', 'gamma')))



    ## Ignoring Whitespace

    test_whitespace_1 = _.passes(
        """|# foo | bar | baz | quux
           |    1 | a   | z   | alpha
           |# this comment will be ignored
           |# so will the following whitespace
           |
           |    2 | b   | y   | beta
           |    3 | c   | x   | gamma""",
        {},
        ("foo", "bar", "baz", "quux"),
        4,
        ((1, 'a', 'z', 'alpha'),
         (2, 'b', 'y', 'beta'),
         (3, 'c', 'x', 'gamma')))




    ## Specifying a custom delimiter

    # You can change delimiters
    test_delimiter_1 = _.passes(
        """|## Delimiter: ,
           |1,2,3
           |4,5,6""",
        {},
        None,
        3,
        ((1, 2, 3),
         (4, 5, 6)))

    # The same delimiter is used for column names as is used for the
    # rows
    test_delimiter_2 = _.passes(
        """|## Delimiter: ,
           |## Column-names: a, b, c
           |1, 2, 3
           |4, 5, 6""",
        {},
        ('a', 'b', 'c'),
        3,
        ((1, 2, 3),
         (4, 5, 6)))

    test_delimiter_3 = _.passes(
        """|## Delimiter: ,
           |# a, b, c
           |1, 2, 3
           |4, 5, 6""",
        {},
        ('a', 'b', 'c'),
        3,
        ((1, 2, 3),
         (4, 5, 6)))

    # Multi-character delimiters are supported (but don't do it, it's
    # dumb)
    test_multichar_delimiter_1 = _.passes(
        """|## Delimiter: --
           |1--2--3
           |4--5--6""",
        {}, None, 3, ((1, 2, 3), (4, 5, 6)))
        


    ## Specifying Headers

    test_header_1 = _.passes(
        """|## Title: sample dataset
           |1|a|z
           |2|b|y""",
        {'title': 'sample dataset'},
        None,
        3,
        ((1, 'a', 'z'),
         (2, 'b', 'y')))

    # header names are case insensitive
    test_header_2 = _.passes(
        """|## TItLE: sample dataset
           |1|a|z
           |2|b|y""",
        {'title': 'sample dataset'},
        None,
        3,
        ((1, 'a', 'z'),
         (2, 'b', 'y')))

    # header values are case-preserving
    test_header_3 = _.passes(
        """|## title: SAMPLE DATASET
           |1|a|z
           |2|b|y""",
        {'title': 'SAMPLE DATASET'},
        None,
        3,
        ((1, 'a', 'z'),
         (2, 'b', 'y')))

    # The three meta-info headers are not passed through. (They'll be
    # recreated and possibly different if the object is re-serialized
    # later.)
    test_header_2 = _.passes(
        """|## Foo: Bar
           |## Delimiter:,
           |## Column-names: a,b,c
           |## Typemap: int, int, int
           |1,2,3
           |4,5,6""",
        {'foo': 'Bar'},
        ('a', 'b', 'c'),
        3,
        ((1, 2, 3), (4, 5, 6)))



    ## Specifying typemaps

    # (If the typemap isn't being used, the inferred type will be int,
    # which will mess things up.)
    test_typemap_1 = _.passes(
        """|## Delimiter:,
           |## Column-names: a,b,c
           |## Typemap: float, float, float
           |1,2,3
           |4.2,5.1,6.05""",
        {},
        ('a', 'b', 'c'),
        3,
        ((1,2,3), (4.2, 5.1, 6.05)))




    ## Specifying column names

    test_colnames_1 = _.passes(
        """|## Delimiter:,
           |## Column-names: a,b,c
           |1,2,3
           |4,5,6""",
        {},
        ('a', 'b', 'c'),
        3,
        ((1,2,3), (4,5,6)))

    test_colnames_2 = _.passes(
        """|## Delimiter:,
           |# a,b,c
           |1,2,3
           |4,5,6""",
        {},
        ('a', 'b', 'c'),
        3,
        ((1,2,3), (4,5,6)))

    test_colnames_3 = _.passes(
        """|## Delimiter:,
           |#a,b,c
           |1,2,3
           |4,5,6""",
        {},
        ('a', 'b', 'c'),
        3,
        ((1,2,3), (4,5,6)))

    test_colnames_4 = _.passes(
        """|## Delimiter:,
           |## a,b,c
           |1,2,3
           |4,5,6""",
        {},
        ('a', 'b', 'c'),
        3,
        ((1,2,3), (4,5,6)))

    test_colnames_5 = _.passes(
        """|## Delimiter:,
           |##a,b,c
           |1,2,3
           |4,5,6""",
        {},
        ('a', 'b', 'c'),
        3,
        ((1,2,3), (4,5,6)))

    # silk-style column names
    test_colnames_6 = _.passes(
        """|a|b|c
           |1|2|3
           |4|5|6""",
        {},
        ('a', 'b', 'c'),
        3,
        ((1,2,3), (4,5,6)),
        first_line_is_colnames=True)

    test_colnames_7 = _.passes(
        """|  a|  b|  c
           |  1|  2|  3
           |  4|  5|  6""",
        {},
        ('a', 'b', 'c'),
        3,
        ((1,2,3), (4,5,6)),
        first_line_is_colnames=True)



    ## Various kinds of empty datasets

    # Nothing
    test_empty_1 = _.passes(
        "", {}, None, 0, tuple())

    # No header, No data
    test_missing_1 = _.passes(
        """|# Comment and no data""",
        {}, None, 0, tuple())

    # No data
    test_missing_2 = _.passes(
        """|## header: yes
           |## data: nope""",
        {'header': 'yes', 'data': 'nope'}, None, 0, tuple())

    # Data, no header
    test_missing_3 = _.passes(
        """|1|3|5
           |2|4|6""",
        {},
        None,
        3,
        ((1,3,5), (2,4,6)))


    # Missing values
    #
    # Missing values are passed on to the type conversion function (as
    # an empty string), which can decide whether and how to deal with
    # them.
    test_missing_4 = _.passes(
        """|#foo|bar|baz
           |1||5
           |2|4|6""",
        {},
        ('foo', 'bar', 'baz'),
        3,
        ((1,'NaN',5), (2,4,6)),
        orig_typemap=(int, _.missing_typemap, int))

    # The default typemaps create a bit of a mess out of this....
    test_missing_5 = _.passes(
        """|#foo|bar|baz
           |1||5
           |2|4|6""",
        {},
        ('foo', 'bar', 'baz'),
        3,
        ((1,'',5), (2,'4',6)))        

    test_missing_6 = _.fails(
        """|#foo|bar|baz
           |1|3|5
           |2||6""",
        TypeConversionException,
        iterate_rows=True)


    
# TODO: to_stream, to_csv_stream

if __name__ == '__main__':
    unittest.main()
