#  Copyright (C) 2012 by Carnegie Mellon University.
#  
#  @OPENSOURCE_HEADER_START@
#  Use of the Rayon and related source code is subject to the terms of
#  the following licenses:
#  
#  GNU Public License (GPL) Rights pursuant to Version 2, June 1991
#  Government Purpose License Rights (GPLR) pursuant to DFARS 252.227.7013
#  
#  NO WARRANTY
#  
#  ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER 
#  PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY 
#  PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN 
#  "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY 
#  KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT 
#  LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE, 
#  MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE 
#  OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT, 
#  SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY 
#  TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF 
#  WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES. 
#  LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF 
#  CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON 
#  CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE 
#  DELIVERABLES UNDER THIS LICENSE.
#  
#  Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie 
#  Mellon University, its trustees, officers, employees, and agents from 
#  all claims or demands made against them (and any related losses, 
#  expenses, or attorney's fees) arising out of, or relating to Licensee's 
#  and/or its sub licensees' negligent use or willful misuse of or 
#  negligent conduct or willful misconduct regarding the Software, 
#  facilities, or other rights or assistance granted by Carnegie Mellon 
#  University under this License, including, but not limited to, any 
#  claims of product liability, personal injury, death, damage to 
#  property, or violation of any laws or regulations.
#  
#  Carnegie Mellon University Software Engineering Institute authored 
#  documents are sponsored by the U.S. Department of Defense under 
#  Contract FA8721-05-C-0003. Carnegie Mellon University retains 
#  copyrights in all material produced under this contract. The U.S. 
#  Government retains a non-exclusive, royalty-free license to publish or 
#  reproduce these documents, or allow others to do so, for U.S. 
#  Government purposes only pursuant to the copyright license under the 
#  contract clause at 252.227.7013.
#  @OPENSOURCE_HEADER_END@

from __future__ import division
import unittest
from rayon.data import Dataset
from rayon.datautils import *
from netsa.data.times import make_datetime

even_elems = [0,10,20,30,40,50,60,70,80,90]
odd_elems = [0,10,20,30,40,50,60,70,80,90,100]

class t_data_functions(unittest.TestCase):
    def test_mean_1(self):
        s = Statistics(even_elems)
        self.assertAlmostEqual(45.0, mean(even_elems), 2)

    def test_mean_2(self):
        self.assertRaises(ValueError, lambda: mean([]))

    # The same tests as in t_statistics, but the function is
    # implemented in a different place (TODO: change this, dummy),
    # so...
    def test_percentile_1(self):
        "non-edge, even number of elements"
        self.assertEqual(50, percentile(even_elems, 50))
        
    def test_percentile_2(self):
        "non-edge, odd number of elements"
        s = Statistics(odd_elems)
        self.assertAlmostEqual(55, percentile(odd_elems, 50), 2)
        
    def test_percentile_3(self):
        "low edge, even number of elements"
        self.assertEqual(0, percentile(even_elems, 0))

    def test_percentile_4(self):
        "low edge, odd number of elements"
        self.assertEqual(0, percentile(odd_elems, 0))

    def test_percentile_5(self):
        "below low edge, even number of elements"
        self.assertEqual(0, percentile(even_elems, -1))

    def test_percentile_6(self):
        "below low edge, odd number of elements"
        self.assertEqual(0, percentile(odd_elems, -1))
    
    def test_percentile_7(self):
        "high edge, even number of elements"
        self.assertEqual(90, percentile(even_elems, 100))
        
    def test_percentile_8(self):
        "high edge, odd number of elements"
        self.assertEqual(100, percentile(odd_elems, 100))
        
    def test_percentile_9(self):
        "above high edge, even number of elements"
        self.assertEqual(90, percentile(even_elems, 100))
        
    def test_percentile_10(self):
        "above high edge, odd number of elements"
        self.assertEqual(100, percentile(odd_elems, 100))

    def test_population_variance_1(self):
        self.assertAlmostEqual(825.0, population_variance(even_elems), 2)

    def test_population_variance_2(self):
        self.assertRaises(ValueError, lambda: population_variance([]))
                
    def test_sample_variance_1(self):
        self.assertAlmostEqual(916.666, sample_variance(even_elems), 2)

    def test_sample_variance_2(self):
        self.assertRaises(ValueError, lambda: sample_variance([]))
                
    def test_sample_variance_3(self):
        self.assertEqual(0.0, sample_variance([1]))
                
    def test_population_standard_deviation_1(self):
        self.assertAlmostEqual(28.72, population_stdev(even_elems), 2)
        
    def test_population_standard_deviation_2(self):
        self.assertAlmostEqual(0.0, population_stdev([1]), 2)

    def test_population_standard_deviation_3(self):
        self.assertRaises(ValueError, lambda: population_stdev([]))

    def test_sample_standard_deviation_1(self):
        self.assertAlmostEqual(30.276, sample_stdev(even_elems), 2)
        
    def test_sample_standard_deviation_2(self):
        self.assertAlmostEqual(0.0, sample_stdev([1]), 2)

    def test_sample_standard_deviation_3(self):
        self.assertRaises(ValueError, lambda: sample_stdev([]))

class t_Statistics(unittest.TestCase):
    def test_percentile_1(self):
        "non-edge, even number of elements"
        s = Statistics(even_elems)
        self.assertEqual(50, s.percentile(50))
        
    def test_percentile_2(self):
        "non-edge, odd number of elements"
        s = Statistics(odd_elems)
        self.assertAlmostEqual(55, s.percentile(50), 2)
        
    def test_percentile_3(self):
        "low edge, even number of elements"
        s = Statistics(even_elems)
        self.assertEqual(0, s.percentile(0))

    def test_percentile_4(self):
        "low edge, odd number of elements"
        s = Statistics(odd_elems)
        self.assertEqual(0, s.percentile(0))

    def test_percentile_5(self):
        "below low edge, even number of elements"
        s = Statistics(even_elems)
        self.assertEqual(0, s.percentile(-1))

    def test_percentile_6(self):
        "below low edge, odd number of elements"
        s = Statistics(odd_elems)
        self.assertEqual(0, s.percentile(-1))
    
    def test_percentile_7(self):
        "high edge, even number of elements"
        s = Statistics(even_elems)
        self.assertEqual(90, s.percentile(100))
        
    def test_percentile_8(self):
        "high edge, odd number of elements"
        s = Statistics(odd_elems)
        self.assertEqual(100, s.percentile(100))
        
    def test_percentile_9(self):
        "above high edge, even number of elements"
        s = Statistics(even_elems)
        self.assertEqual(90, s.percentile(100))
        
    def test_percentile_10(self):
        "above high edge, odd number of elements"
        s = Statistics(odd_elems)
        self.assertEqual(100, s.percentile(100))
        
    def test_mean_1(self):
        s = Statistics(even_elems)
        self.assertAlmostEqual(45.0, s.mean(), 2)

    def test_mean_2(self):
        s = Statistics([])
        self.assertRaises(ValueError, lambda: s.mean())

    def test_population_variance_1(self):
        s = Statistics(even_elems)
        self.assertAlmostEqual(825.0, s.variance(), 2)

    def test_population_variance_2(self):
        s = Statistics([])
        self.assertRaises(ValueError, lambda: s.variance())
                
    def test_sample_variance_1(self):
        s = Statistics(even_elems)
        self.assertAlmostEqual(916.666, s.sample_variance(), 2)

    def test_sample_variance_2(self):
        s = Statistics([])
        self.assertRaises(ValueError, lambda: s.sample_variance())
                
    def test_sample_variance_3(self):
        s = Statistics([1])
        self.assertEqual(0.0, s.sample_variance())
                
    def test_population_standard_deviation_1(self):
        s = Statistics(even_elems)
        self.assertAlmostEqual(28.72, s.stdev(), 2)
        
    def test_population_standard_deviation_2(self):
        s = Statistics([1])
        self.assertAlmostEqual(0.0, s.stdev(), 2)

    def test_population_standard_deviation_3(self):
        s = Statistics([])
        self.assertRaises(ValueError, lambda: s.stdev())

    def test_sample_standard_deviation_1(self):
        s = Statistics(even_elems)
        self.assertAlmostEqual(30.276, s.sample_stdev(), 2)
        
    def test_sample_standard_deviation_2(self):
        s = Statistics([1])
        self.assertAlmostEqual(0.0, s.sample_stdev(), 2)

    def test_sample_standard_deviation_3(self):
        s = Statistics([])
        self.assertRaises(ValueError, lambda: s.sample_stdev())



# Moving calculations

import calendar
def as_seconds(d): 
    return calendar.timegm(d.timetuple())

class t_window_timeseries(unittest.TestCase):
    def test(self):
        times = [make_datetime(x) for x in [0, 100, 200, 300, 400, 500, 600]]
        vals  = [0, 1, 2, 3, 4, 5, 6]
        identity = lambda times, vals: vals
        new_times, new_vals = window_timeseries(times, vals, identity) 
        times_as_secs = [as_seconds(d) for d in new_times]
        self.assertEqual([[150, 250, 350, 450], 
                          ([0,1,2,3], [1,2,3,4], [2,3,4,5], [3,4,5,6])],
                         [times_as_secs, new_vals])
            

class t_moving_average_raw(unittest.TestCase):
    def test(self):
        times = [make_datetime(x) for x in [0, 100, 200, 300, 400, 500, 600]]
        vals  = [0, 1, 2, 3, 4, 5, 6]
        expected_out = [tuple(make_datetime(x) for x in [150, 250, 350, 450]),
                        (1.5, 2.5, 3.5, 4.5)]
        self.assertEqual(expected_out, moving_average_raw(times, vals))

class t_moving_average(unittest.TestCase):
    def test(self):
        times = [make_datetime(x) for x in [0, 100, 200, 300, 400, 500, 600]]
        vals  = [0, 1, 2, 3, 4, 5, 6]
        in_data = Dataset.from_columns([times, vals])
        expected_out = [tuple(make_datetime(x) for x in [150, 250, 350, 450]),
                        (1.5, 2.5, 3.5, 4.5)]
        expected_out_data = Dataset.from_columns(
            expected_out,
            colnames=['time', 'value'])
        self.assertEqual(expected_out_data, moving_average(in_data))
        

if __name__ == '__main__':
    unittest.main()


