#  Copyright (C) 2012 by Carnegie Mellon University.
#  
#  @OPENSOURCE_HEADER_START@
#  Use of the Rayon and related source code is subject to the terms of
#  the following licenses:
#  
#  GNU Public License (GPL) Rights pursuant to Version 2, June 1991
#  Government Purpose License Rights (GPLR) pursuant to DFARS 252.227.7013
#  
#  NO WARRANTY
#  
#  ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER 
#  PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY 
#  PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN 
#  "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY 
#  KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT 
#  LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE, 
#  MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE 
#  OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT, 
#  SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY 
#  TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF 
#  WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES. 
#  LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF 
#  CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON 
#  CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE 
#  DELIVERABLES UNDER THIS LICENSE.
#  
#  Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie 
#  Mellon University, its trustees, officers, employees, and agents from 
#  all claims or demands made against them (and any related losses, 
#  expenses, or attorney's fees) arising out of, or relating to Licensee's 
#  and/or its sub licensees' negligent use or willful misuse of or 
#  negligent conduct or willful misconduct regarding the Software, 
#  facilities, or other rights or assistance granted by Carnegie Mellon 
#  University under this License, including, but not limited to, any 
#  claims of product liability, personal injury, death, damage to 
#  property, or violation of any laws or regulations.
#  
#  Carnegie Mellon University Software Engineering Institute authored 
#  documents are sponsored by the U.S. Department of Defense under 
#  Contract FA8721-05-C-0003. Carnegie Mellon University retains 
#  copyrights in all material produced under this contract. The U.S. 
#  Government retains a non-exclusive, royalty-free license to publish or 
#  reproduce these documents, or allow others to do so, for U.S. 
#  Government purposes only pursuant to the copyright license under the 
#  contract clause at 252.227.7013.
#  @OPENSOURCE_HEADER_END@
from __future__ import division

import unittest
import itertools

from rayon.geometry import Rectangle, ScreenPoint

from rayon.subdivide import Subdivision, subdivide_htiles, subdivide_vtiles

def check_stream_size(stream, size):
    # Returns 0 if the (potentially infinite) stream terminates at the
    # given size, otherwise, -1 if it terminates early, 1 if it
    # terminates late
    if len(items) < size:    return -1
    elif len(items) == size: return 0
    else:                    return 1
                 

def as_tuple(stream, expected_size):
    # Returns the stream as a list of (ideally) the expected_size. If
    # the number of items in the stream is <= expected_size, it
    # returns a list of all the items. If the number of items is >
    # expected_size, it returns one more than the expected number of
    # items.
    rc = zip(*list(itertools.izip(xrange(expected_size + 1), stream)))
    try:
        return tuple(rc[1])
    except IndexError:
        # Empty list
        return []

class t_Subdivision(unittest.TestCase):
    def test_access_1(self):
        r1 = Rectangle(100, 50)
        pt1 = ScreenPoint(0, 0)
        sd = Subdivision(r1, pt1)
        r2, pt2 = sd
        self.assertEqual((r1, pt1), (r2, pt2))
    def test_types_1(self):
        self.assertRaises(TypeError,
                          lambda: Subdivision(Rectangle(100, 50), (0, 0)))
    def test_types_2(self):
        self.assertRaises(TypeError,
                          lambda: Subdivision((100, 50), ScreenPoint(0, 0)))
    def test_eq_1(self):
        r1, pt1 = Rectangle(100, 50), ScreenPoint(0, 0)
        r2, pt2 = Rectangle(100, 50), ScreenPoint(0, 0)
        self.assertEqual(Subdivision(r1, pt1), Subdivision(r2, pt2))
    def test_neq_1(self):
        r1, pt1 = Rectangle(100, 50), ScreenPoint(0, 0)
        r2, pt2 = Rectangle(50, 50), ScreenPoint(0, 0)
        self.assertNotEqual(Subdivision(r1, pt1), Subdivision(r2, pt2))
    def test_neq_2(self):
        r1, pt1 = Rectangle(100, 50), ScreenPoint(0, 0)
        r2, pt2 = Rectangle(100, 50), ScreenPoint(0, 10)
        self.assertNotEqual(Subdivision(r1, pt1), Subdivision(r2, pt2))
    def test_neq_3(self):
        r1, pt1 = Rectangle(100, 50), ScreenPoint(0, 0)
        r2, pt2 = Rectangle(100, 50), ScreenPoint(0, 10)
        self.assertNotEqual(Subdivision(r1, pt1), (r2, pt2))

def subd(w, h, x, y):
    return Subdivision(Rectangle(w, h), ScreenPoint(x, y))


def rotate_grid(cells, w, h):
    # Example input:
    #  [0, 1, 2, 3,
    #   4, 5, 6, 7,
    #   8, 9,10,11]
    #
    # Example output:
    #  [0, 4, 8,
    #   1, 5, 9,
    #   2, 6,10,
    #   3, 7,11]
    if w == 0 or h == 0:
        return cells
    else:
        return tuple(cells[row * w + col]
                     for col in xrange(w)
                     for row in xrange(h))
    

def sd_tiles_test(inrect_width,
                  inrect_height,
                  col_weights,
                  row_weights,
                  subdivs):
    subdivs = tuple(subdivs)
    def num_cells(weights):
        if len(weights) == 0: return 1
        else: return len(weights)

    inrect = Rectangle(inrect_width, inrect_height)

    def t1(self):
        h_subdivs = as_tuple(subdivide_htiles(inrect, col_weights, row_weights),
                             num_cells(col_weights) * num_cells(row_weights))
        self.assertEqual(h_subdivs, subdivs)
        

    def t2(self):
        v_subdivs = as_tuple(subdivide_vtiles(inrect, col_weights, row_weights),
                             num_cells(col_weights) * num_cells(row_weights))
        rotated = rotate_grid(subdivs, len(col_weights), len(row_weights))
        from pprint import pprint
        self.assertEqual(v_subdivs, rotate_grid(subdivs,
                                                len(col_weights),
                                                len(row_weights)))
                         
    return t1, t2





class t_subdivide_htiles(unittest.TestCase):
    (test_cols_h_1, test_cols_v_1) = sd_tiles_test(
        400, 50, [3,1], [],
        [subd(300, 50, 0, 0),
         subd(100, 50, 300, 0)])
    
    (test_cols_h_2, test_cols_v_2) = sd_tiles_test(
        400, 50, [1,1,1,1], [],
        [subd(100, 50, 0, 0),
         subd(100, 50, 100, 0),
         subd(100, 50, 200, 0),
         subd(100, 50, 300, 0)])
    
    (test_cols_h_2, test_cols_v_2) = sd_tiles_test(
        400, 50, [1,1,1,1], [],
        [subd(100, 50, 0, 0),
         subd(100, 50, 100, 0),
         subd(100, 50, 200, 0),
         subd(100, 50, 300, 0)])
    
    (test_cols_h_3, test_cols_v_3) = sd_tiles_test(
        400, 50, [1,2,1], [],
        [subd(100, 50, 0, 0),
         subd(200, 50, 100, 0),
         subd(100, 50, 300, 0)])
    
    (test_cols_h_4, test_cols_v_4) = sd_tiles_test(
        600, 50, [1,2,3], [],
        [subd(100, 50, 0, 0),
         subd(200, 50, 100, 0),
         subd(300, 50, 300, 0)])
    
    (test_rows_h_1, test_rows_v_1) = sd_tiles_test(
        50, 400, [], [3,1],
        [subd(50, 300, 0, 0),
         subd(50, 100, 0, 300)])

    (test_rows_h_2, test_rows_v_2) = sd_tiles_test(
        50, 400, [], [1,1,1,1],
        [subd(50, 100, 0, 0),
         subd(50, 100, 0, 100),
         subd(50, 100, 0, 200),
         subd(50, 100, 0, 300)])
    
    (test_rows_h_3, test_rows_v_3) = sd_tiles_test(
        50, 400, [], [1,2,1],
        [subd(50, 100, 0, 0),
         subd(50, 200, 0, 100),
         subd(50, 100, 0, 300)])

    (test_rows_h_4, test_rows_v_4) = sd_tiles_test(
        50, 600, [], [1,2,3],
        [subd(50, 100, 0, 0),
         subd(50, 200, 0, 100),
         subd(50, 300, 0, 300)])

    (test_both_h_1, test_both_v_1) = sd_tiles_test(
        400, 400, [1,1,1,1], [1,1,1,1],
        # row 1
        [subd(100, 100, 0, 0),
         subd(100, 100, 100, 0),
         subd(100, 100, 200, 0),
         subd(100, 100, 300, 0),
         # row 2
         subd(100, 100, 0  , 100),
         subd(100, 100, 100, 100),
         subd(100, 100, 200, 100),
         subd(100, 100, 300, 100),
         # row 3
         subd(100, 100, 0  , 200),
         subd(100, 100, 100, 200),
         subd(100, 100, 200, 200),
         subd(100, 100, 300, 200),
         # row 4
         subd(100, 100, 0  , 300),
         subd(100, 100, 100, 300),
         subd(100, 100, 200, 300),
         subd(100, 100, 300, 300)])

    (test_both_h_2, test_both_v_2) = sd_tiles_test(
        400, 400, [1,1,1,1], [1,2,1],
        # row 1
        [subd(100, 100, 0, 0),
         subd(100, 100, 100, 0),
         subd(100, 100, 200, 0),
         subd(100, 100, 300, 0),
         # row 2
         subd(100, 200, 0  , 100),
         subd(100, 200, 100, 100),
         subd(100, 200, 200, 100),
         subd(100, 200, 300, 100),
         # row 3
         subd(100, 100, 0  , 300),
         subd(100, 100, 100, 300),
         subd(100, 100, 200, 300),
         subd(100, 100, 300, 300)])

    (test_both_h_4, test_both_v_4) = sd_tiles_test(
        400, 600, [1,1,1,1], [1,2,3],
        # row 1
        [subd(100, 100, 0, 0),
         subd(100, 100, 100, 0),
         subd(100, 100, 200, 0),
         subd(100, 100, 300, 0),
         # row 2
         subd(100, 200, 0  , 100),
         subd(100, 200, 100, 100),
         subd(100, 200, 200, 100),
         subd(100, 200, 300, 100),
         # row 3
         subd(100, 300, 0  , 300),
         subd(100, 300, 100, 300),
         subd(100, 300, 200, 300),
         subd(100, 300, 300, 300)])

    # Odds and ends

    # If one of the rectangle dimensions is zero, the subdivided
    # rectangles will use zero as that dimension, and will position
    # themselves at zero in x (for width) or y (for height). These
    # don't make any practical sense, but it's the calling code's job
    # to decide of a given rectangle is usable for its purposes. This
    # code is only the messenger.
    #
    # The important thing is that the number of cells implied by the
    # weights always exists. If they ask for 3 columns and 3 rows,
    # they get 9 cells, doggonit.
    (test_zero_width_h_1, test_zero_width_v_1) =  sd_tiles_test(
        0, 50, [3,1], [],
        [subd(0, 50, 0, 0),
         subd(0, 50, 0, 0)])
    
    (test_zero_width_h_2, test_zero_width_v_2) = sd_tiles_test(
        0, 400, [], [3,1],
        [subd(0, 300, 0, 0),
         subd(0, 100, 0, 300)])

    (test_zero_width_h_3, test_zero_width_v_3) = sd_tiles_test(
        0, 400, [1,1], [3,1],
        [subd(0, 300, 0, 0),
         subd(0, 300, 0, 0),
         subd(0, 100, 0, 300),
         subd(0, 100, 0, 300)])

    (test_zero_height_h_1, test_zero_height_v_1) = sd_tiles_test(
        400, 0, [3,1], [],
        [subd(300, 0, 0, 0),
         subd(100, 0, 300, 0)])

    (test_zero_height_h_2, test_zero_height_v_2) = sd_tiles_test(
        400, 0, [], [3,1],
        [subd(400, 0, 0, 0),
         subd(400, 0, 0, 0)])


    (test_zero_height_h_3, test_zero_height_v_3) = sd_tiles_test(
        400, 0, [1,1], [3,1],
        [subd(200, 0, 0, 0),
         subd(200, 0, 200, 0),
         subd(200, 0, 0, 0),
         subd(200, 0, 200, 0)])

    # A 0x0 input rect will always output a stream of 0x0 subrects
    # anchored at (0,0), as many as there should be cells according to
    # the weights.
    (test_zero_width_height_h_1, test_zero_width_height_v_1) = sd_tiles_test(
        0, 0, [1,1], [],
        [subd(0, 0, 0, 0),
         subd(0, 0, 0, 0),])

    (test_zero_width_height_h_2, test_zero_width_height_v_2) = sd_tiles_test(
        0, 0, [], [1,1],
        [subd(0, 0, 0, 0),
         subd(0, 0, 0, 0),])

    (test_zero_width_height_h_3, test_zero_width_height_v_3) = sd_tiles_test(
        0, 0, [1,1], [1,1],
        [subd(0, 0, 0, 0),
         subd(0, 0, 0, 0),
         subd(0, 0, 0, 0),
         subd(0, 0, 0, 0)])

    # Weights of 0 produce zero-dimension rectangles, but they are
    # produced. If you ask for columns of [1,0,1] and rows of [1,1,1],
    # you get 9 subdivisions. You even get 9 subdivisions if you ask
    # for [0,0,0] and [0,0,0]

    (test_zero_weights_h_1, test_zero_weights_v_1) = sd_tiles_test(
        400, 50, [0, 1], [],
        [subd(0, 50, 0, 0),
         subd(400, 50, 0, 0)])

    (test_zero_weights_h_2, test_zero_weights_v_2) = sd_tiles_test(
        400, 50, [], [0, 1],
        [subd(400, 0, 0, 0),
         subd(400, 50, 0, 0)])

    

    # A weight of [1] (or any single number) is equivalent to a weight
    # of [].
    (test_single_weight_h_1, test_single_weight_v_1) = sd_tiles_test(
        400, 50, [1], [1,1],
        [subd(400, 25, 0, 0),
         subd(400, 25, 0, 25)])
    
    (test_single_weight_h_2, test_single_weight_v_2) = sd_tiles_test(
        400, 50, [1,1], [1],
        [subd(200, 50, 0, 0),
         subd(200, 50, 200, 0)])

    # Non-integer weights work. They're dumb, though, don't use them.
    # ([.25, .50] is identical to [1, 2], it just requires floating
    # point math, may round in stupid ways, and smells funny.)

    test_fractional_weights = sd_tiles_test(
        300, 50, [.25, .50], [.5, .5],
        [subd(100, 25, 0, 0),
         subd(200, 25, 100, 0),
         subd(100, 25, 0, 25),
         subd(200, 25, 100, 25)])

    # Fractional widths work, too, and may even be somewhat sensible
    # -- say, if you're scaling everything to be between 0 and 1. It
    # still smells funny, though.
    
    (test_fractional_widths_h_1, test_fractional_widths_v_1) = sd_tiles_test(
        1, 1, [2,2,1], [1,1],
        [subd(2/5, 1/2, 0, 0),
         subd(2/5, 1/2, 2/5, 0),
         subd(1/5, 1/2, 4/5, 0),
         subd(2/5, 1/2, 0, 1/2),
         subd(2/5, 1/2, 2/5, 1/2),
         subd(1/5, 1/2, 4/5, 1/2)])
    
    (test_fractional_widths_h_2, test_fractional_widths_v_2) = sd_tiles_test(
        1/2, 1/2, [2,2,1], [1,1],
        [subd(2/10, 1/4, 0, 0),
         subd(2/10, 1/4, 2/10, 0),
         subd(1/10, 1/4, 4/10, 0),
         subd(2/10, 1/4, 0, 1/4),
         subd(2/10, 1/4, 2/10, 1/4),
         subd(1/10, 1/4, 4/10, 1/4)])
                                               
if __name__ == "__main__":
    unittest.main()
