import rayon.scales
import unittest

from netsa.data.times import make_datetime, make_timedelta


class t_LinearBinnedScale(unittest.TestCase):

    def setUp(self):
        self.s1 = rayon.scales.LinearBinnedScale(
            input_min=50, input_max=90, bin_size=10, bin_start=0)
        self.s2 = rayon.scales.LinearBinnedScale(
            input_min=50, input_max=90,
            output_min=1, output_max=2,
            bin_size=10, bin_start=0)
        self.s3 = rayon.scales.LinearBinnedScale(
            input_min=-50, input_max=40, bin_size=10, bin_start=0)
        self.s4 = rayon.scales.LinearBinnedScale(bin_size=10, bin_start=0)
        self.s5 = rayon.scales.LinearBinnedScale(
            input_min=52, input_max=92, bin_size=10, bin_start=2)

    def assertTuplesAlmostEqual(self, t1, t2):
        self.assertEquals(len(t1), len(t2))
        for i1, i2 in zip(t1, t2):
            self.assertAlmostEqual(i1, i2)

    def test_call_1(self):
        self.assertTuplesAlmostEqual(self.s1(50), (0.0, 0.2))

    def test_call_3(self):
        self.assertTuplesAlmostEqual(self.s1(70), (0.4, 0.6))

    def test_call_4(self):
        self.assertTuplesAlmostEqual(self.s2(50), (1.0, 1.2))

    def test_call_5(self):
        self.assertTuplesAlmostEqual(self.s2(90), (1.8, 2.0))

    def test_call_6(self):
        self.assertTuplesAlmostEqual(self.s2(70), (1.4, 1.6))

    def test_call_7(self):
        self.assertTuplesAlmostEqual(self.s3(-50), (0.0, 0.1))

    def test_call_8(self):
        self.assertTuplesAlmostEqual(self.s3(40), (0.9, 1.0))

    def test_call_9(self):
        self.assertTuplesAlmostEqual(self.s3(0), (.5, .6))

    def test_call_10(self):
        self.assertTuplesAlmostEqual(self.s5(52), (0.0, 0.2))

    def test_bad_limits_1(self):
        def check():
            rayon.scales.LinearBinnedScale(input_min=2, input_max=30,
                                           bin_size=10, bin_start=0)
        self.assertRaises(rayon.scales.RayonLimitException, check)

    def test_bad_limits_2(self):
        def check():
            rayon.scales.LinearBinnedScale(input_min=10, input_max=31,
                                           bin_size=10, bin_start=0)
        self.assertRaises(rayon.scales.RayonLimitException, check)

    def test_bad_limits_3(self):
        def check():
            rayon.scales.LinearBinnedScale(input_min=10, input_max=30.1,
                                           bin_size=10, bin_start=0)
        self.assertRaises(rayon.scales.RayonLimitException, check)

    def test_bad_input_1(self):
        def check():
            self.s1(12)
        self.assertRaises(rayon.scales.RayonRangeException, check)

    def test_bad_input_2(self):
        def check():
            self.s1(52)
        self.assertRaises(rayon.scales.RayonRangeException, check)

    def test_bad_input_3(self):
        def check():
            self.s1(102)
        self.assertRaises(rayon.scales.RayonRangeException, check)


    def test_attach_1(self):
        self.s4.attach((0, 10, 30, 90), "owner", 1)
        self.assertTuplesAlmostEqual(self.s4(0), (0.0, 0.1))

    def test_attach_2(self):
        self.s4.attach((0, 10, 30, 90), "owner", 1)
        self.assertTuplesAlmostEqual(self.s4(90), (0.9, 1.0))

    def test_attach_3(self):
        self.s4.attach((90,), "owner", 1)
        self.assertTuplesAlmostEqual(self.s4(90), (0.0, 1.0))
        def check():
            self.s4(0)
        self.assertRaises(rayon.scales.RayonRangeException, check)

    def test_attach_4(self):
        self.s4.attach((0, 10, 30, 90), "owner", 1)
        self.assertTuplesAlmostEqual(self.s4(90), (0.9, 1.0))
        self.s4.attach((20,), "owner", 2)
        self.assertTuplesAlmostEqual(self.s4(90), (0.9, 1.0))
        self.s4.attach((190,), "owner", 3)
        self.assertTuplesAlmostEqual(self.s4(90), (0.45, .5))

    def test_detach_1(self):
        self.s4.attach((0, 10, 30, 90), "owner", 1)
        self.assertTuplesAlmostEqual(self.s4(90), (0.9, 1.0))
        self.s4.attach((190,), "owner", 3)
        self.assertTuplesAlmostEqual(self.s4(90), (0.45, .5))
        self.s4.detach("owner", 3)
        self.assertTuplesAlmostEqual(self.s4(90), (0.9, 1.0))
        

    def test_bad_attach_1(self):
        def check():
            self.s4.attach((0, 15, 30, 90), "owner")
            self.s4.get_input_min()
        self.assertRaises(rayon.scales.RayonRangeException, check)


class t_TimeBinnedScale(unittest.TestCase):

    def setUp(self):
        # Use the default bin_start of the unix epoch
        self.s1 = rayon.scales.TimeBinnedScale(
            input_min=make_datetime("2001-01-01 00:00:00"),
            input_max=make_datetime("2001-01-01 21:00:00"),
            bin_size=make_timedelta("PT3H"))
        self.s2 = rayon.scales.TimeBinnedScale(
            input_min=make_datetime("2001-01-01 00:00:00"),
            input_max=make_datetime("2001-01-01 21:00:00"),
            output_min=1, output_max=2,
            bin_size=make_timedelta("PT3H"))
        self.s3 = rayon.scales.TimeBinnedScale(
            bin_size=make_timedelta("PT3H"))
        self.s4 = rayon.scales.TimeBinnedScale(
            bin_size=make_timedelta("PT3H"),
            bin_start=make_datetime("2000-01-01 00:30:00"))

    def assertTuplesAlmostEqual(self, t1, t2):
        self.assertEquals(len(t1), len(t2))
        for i1, i2 in zip(t1, t2):
            self.assertAlmostEqual(i1, i2)

    def test_call_1(self):
        dt = make_datetime("2001-01-01 00:00:00")
        self.assertTuplesAlmostEqual(self.s1(dt), (0.0, 0.125))

    def test_call_2(self):
        dt = make_datetime("2001-01-01 21:00:00")
        self.assertTuplesAlmostEqual(self.s1(dt), (0.875, 1.0))

    def test_call_3(self):
        dt = make_datetime("2001-01-01 00:00:00")
        self.assertTuplesAlmostEqual(self.s2(dt), (1.0, 1.125))

    def test_call_4(self):
        dt = make_datetime("2001-01-01 21:00:00")
        self.assertTuplesAlmostEqual(self.s2(dt), (1.875, 2.0))


    def test_bad_call_1(self):
        dt = make_datetime("2001-01-02 00:00:00")
        def check():
            self.s1(dt)
        self.assertRaises(rayon.scales.RayonRangeException, check)
        
    def test_bad_call_2(self):
        dt = make_datetime("2000-01-01 00:00:00")
        def check():
            self.s1(dt)
        self.assertRaises(rayon.scales.RayonRangeException, check)

    def test_bad_limits_1(self):
        def check():
            rayon.scales.TimeBinnedScale(
                input_min=make_datetime("2001-01-01 01:00:00"),
                input_max=make_datetime("2001-01-01 21:00:00"),
                bin_size=make_timedelta("PT3H"))
        self.assertRaises(rayon.scales.RayonLimitException, check)

    def test_bad_limits_2(self):
        def check():
            rayon.scales.TimeBinnedScale(
                input_min=make_datetime("2001-01-01 00:00:00"),
                input_max=make_datetime("2001-01-01 19:00:00"),
                bin_size=make_timedelta("PT3H"))
        self.assertRaises(rayon.scales.RayonLimitException, check)


    def test_attach_1(self):
        self.s3.attach((make_datetime("2001-01-02 00:00:00"),
                        make_datetime("2001-01-02 03:00:00")),
                       "owner", 1)
        dt = make_datetime("2001-01-02 03:00:00")
        self.assertTuplesAlmostEqual(self.s3(dt), (0.5, 1.0))

    def test_attach_2(self):
        self.s3.attach((make_datetime("2001-01-02 00:00:00"),
                        make_datetime("2001-01-02 03:00:00")),
                       "owner", 1)
        dt = make_datetime("2001-01-02 03:00:00")
        self.assertTuplesAlmostEqual(self.s3(dt), (0.5, 1.0))

    def test_detach_1(self):
        self.s3.attach((make_datetime("2001-01-02 00:00:00"),
                        make_datetime("2001-01-02 03:00:00")),
                       "owner", 1)
        dt = make_datetime("2001-01-02 03:00:00")
        self.assertTuplesAlmostEqual(self.s3(dt), (0.5, 1.0))

        self.s3.attach((make_datetime("2001-01-02 09:00:00"),),
                       "owner", 2)
        self.assertTuplesAlmostEqual(self.s3(dt), (0.25, .5))

        self.s3.detach("owner", 2)
        self.assertTuplesAlmostEqual(self.s3(dt), (0.5, 1.0))

    def test_bad_attach_1(self):
        def check():
            self.s3.attach((make_datetime("2001-01-02 00:00:00"),
                            make_datetime("2001-01-02 04:00:00")),
                           "owner", 1)
            self.s3.get_input_min()
        self.assertRaises(rayon.scales.RayonRangeException, check)
            

        

if __name__ == "__main__":
    unittest.main()


__all__ = """
LinearBinnedScale
TimeBinnedScale
""".split()
